#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/03 00:27
# @Author  : Wu_RH
# @FileName: game.py
from ctypes import pointer
from enum import Enum, Flag
import queue
from re import A
import threading
import time
from typing import Any, Union, Callable, List

from ...abs.Lrule import Rule0R
from concurrent.futures import ThreadPoolExecutor, as_completed

from ...abs.Mrule import AbstractMinesValue
from ...abs.Rrule import AbstractClueValue, ValueQuess
from ...abs.board import AbstractBoard
from ...abs.board import AbstractPosition
from . import Summon
from .solver import solver_by_csp, hint_by_csp, Switch, deduced_by_csp, solver_model
from ...utils.impl_obj import MINES_TAG, VALUE_QUESS, POSITION_TAG
from ...utils.tool import get_logger, get_random

from minesweepervariants.config.config import DEFAULT_CONFIG

# ==== 获取默认值 ====
CONFIG = {}
CONFIG.update(DEFAULT_CONFIG)


class Mode(Enum):
    NORMAL = 0  # 普通模式
    EXPERT = 1  # 专家模式
    ULTIMATE = 2  # 终极模式
    PUZZLE = 3  # 纸笔模式(用于调试)


class UMode(Flag):
    ULTIMATE_A = 1
    ULTIMATE_F = 2
    ULTIMATE_S = 4
    ULTIMATE_R = 8
    ULTIMATE_P = 16


NORMAL = Mode.NORMAL
EXPERT = Mode.EXPERT
ULTIMATE = Mode.ULTIMATE
PUZZLE = Mode.PUZZLE

ULTIMATE_A = UMode.ULTIMATE_A
ULTIMATE_F = UMode.ULTIMATE_F
ULTIMATE_S = UMode.ULTIMATE_S
ULTIMATE_R = UMode.ULTIMATE_R
ULTIMATE_P = UMode.ULTIMATE_P


class ValueAsterisk(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        pass

    def __repr__(self) -> str:
        return "*"

    @classmethod
    def type(cls) -> bytes:
        return b"*"

    def code(self) -> bytes:
        return b""

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        return []


class MinesAsterisk(AbstractMinesValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        pass

    def __repr__(self) -> str:
        return "#"

    @classmethod
    def type(cls) -> bytes:
        return b"F*"

    def code(self) -> bytes:
        return b""

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        return []


# VALUE_TAG = ValueAsterisk(POSITION_TAG)
# VALUE_MINES = MinesAsterisk(POSITION_TAG)


class Manger:
    def __init__(self, fn):
        self.fn: Callable = fn
        self.wait_args = None
        self.lock = threading.Lock()
        self.fn_lock = threading.Lock()

    def wait(self):
        my_id = threading.get_ident()  # 当前线程唯一ID

        if not self.fn_lock.locked():
            with self.fn_lock:
                self.fn(thread=True)
            return
        with self.lock:
            self.wait_args = my_id
        while True:
            if self.wait_args is not my_id:
                return
            if not self.fn_lock.locked():
                break
            time.sleep(0.1)
        with self.fn_lock:
            self.fn(thread=True)


class GameSession:
    flag_tag: Any
    clue_tag: Any

    def __init__(
            self, summon: Summon = None, mode=NORMAL, drop_r=False,
            ultimate_mode=ULTIMATE_A | ULTIMATE_F | ULTIMATE_S,
    ):
        self.logger = get_logger()
        self.summon = summon
        self.drop_r = drop_r
        if mode == ULTIMATE:
            if ultimate_mode & ULTIMATE_R:
                self.drop_r = False
            else:
                self.drop_r = True

        self.answer_board = None
        self.board = None

        self.mode = mode
        self.ultimate_mode = ultimate_mode

        self.last_deduced = [None, []]
        self.last_hint = [None, {}]
        self.deducedManger = Manger(self.deduced)
        self.hintManger = Manger(self.hint)
        self.deduced_queue = queue.Queue(maxsize=1)
        self.hint_queue = queue.Queue(maxsize=1)

        self.flag_tag = MinesAsterisk(POSITION_TAG)
        self.clue_tag = ValueAsterisk(POSITION_TAG)

    @property
    def answer_board(self):
        result = self.__dict__["answer_board"]
        if result is None:
            raise AttributeError("answer_board未初始化")
        return result

    @answer_board.setter
    def answer_board(self, value):
        self.__dict__["answer_board"] = value

    @property
    def board(self):
        result = self.__dict__["board"]
        if result is None:
            raise AttributeError("board未初始化")
        return result

    @board.setter
    def board(self, value):
        self.__dict__["board"] = value

    def unbelievable(self, pos, action: int):
        """
        action
            0: 左键点击/翻开/设置非雷
            1: 右键点击/标雷/设置必雷
        """
        if self.answer_board.get_type(pos) == "F" and action == 0:
            return [
                pos for key in self.answer_board.get_board_keys()
                for pos, _ in self.answer_board(key=key)
                if (self.board.get_type(pos) == "N" and
                    self.answer_board.get_type(pos) == "F")
            ]
        if self.answer_board.get_type(pos) == "C" and action == 1:
            return [
                pos for key in self.answer_board.get_board_keys()
                for pos, _ in self.answer_board(key=key)
                if (self.board.get_type(pos) == "N" and
                    self.answer_board.get_type(pos) == "F")
            ]
        all_rules = self.summon.mines_rules.rules[:]
        all_rules += [self.summon.clue_rule, self.summon.mines_clue_rule]
        if self.drop_r:
            all_rules = [rule for rule in all_rules if not isinstance(rule, Rule0R)]
        board = self.board.clone()
        if action == 0:
            board[pos] = self.flag_tag
        else:
            board[pos] = self.clue_tag
        board: AbstractBoard
        print("=" * 20)
        print(board)
        print(board)
        print("=" * 20)
        model = board.get_model()
        switch = Switch()
        for rule in all_rules:
            rule.create_constraints(board, switch)
        for key in board.get_board_keys():
            for pos, obj in board(key=key):
                obj_type = board.get_type(pos)
                var = board.get_variable(pos)
                if obj_type == "F":
                    model.Add(var == 1)
                if obj_type == "C":
                    model.Add(var == 0)
                if obj in [
                    None, MINES_TAG, VALUE_QUESS,
                    self.clue_tag, self.flag_tag
                ]:
                    continue
                obj.create_constraints(board, switch)
        model.AddBoolAnd(switch.get_all_vars())
        state, solver = solver_model(model, True)
        print("unbelievable state:", state)
        if not state:
            return None
        mines_list = []
        for key in board.get_board_keys():
            for pos, var in board(mode="var", key=key):
                if self.board.get_type(pos) != "N":
                    continue
                if solver.Value(var) == 0:
                    continue
                mines_list.append(pos)
        return mines_list

    def thread_hint(self):
        threading.Thread(target=self.hintManger.wait).start()

    def thread_deduced(self):
        threading.Thread(target=self.deducedManger.wait).start()

    def create_board(self) -> Union["AbstractBoard", None]:
        """
        一层具象
        终极模式的规则是 直到推无可推再给下一步线索 如果倒过来想呢
        倒过来就是删无可删 再重头删一遍

        具体操作
        初始化: 将所有雷设置为None
        第一步: 将整个版面的线索都替换为雷试下能不能无解 如果无解代表矛盾
        第二步: 现在整个题板都遍历完了一遍 看起来已经是删无可删了 那么就
        """
        board = self.answer_board.clone()
        for rule in (self.summon.mines_rules.rules
                     + [self.summon.clue_rule,
                        self.summon.mines_clue_rule]):
            rule.init_clear(board)
        clues = [i for i in board("CF")]
        print("game init:", board.show_board(), clues)
        get_random().shuffle(clues)
        r_flag = True
        while clues:
            if r_flag and self.drop_r:
                r_flag = solver_by_csp(
                    self.summon.mines_rules,
                    self.summon.clue_rule,
                    self.summon.mines_clue_rule,
                    board.clone(),
                    answer_board=self.answer_board,
                    drop_r=not r_flag
                ) == 1
            while True:
                if not clues:
                    break
                pos, clue = clues.pop()
                if board.get_type(pos) == "C":
                    board.set_value(pos, MINES_TAG)
                elif board.get_type(pos) == "F":
                    board.set_value(pos, self.clue_tag)
                if solver_by_csp(
                        self.summon.mines_rules,
                        self.summon.clue_rule,
                        self.summon.mines_clue_rule,
                        board.clone(), drop_r=not r_flag) == 0:
                    board.set_value(pos, None)
                    break
                board.set_value(pos, clue)
        if r_flag and self.drop_r:
            return None
        self.board = board
        return board

    def chord_clue(self, clue_pos: AbstractPosition) -> list[AbstractPosition]:
        # 看最后一次提示有没有包含该格的单线索
        VALUE = self.board.get_config(clue_pos.board_key, "VALUE")
        MINES = self.board.get_config(clue_pos.board_key, "MINES")
        self.logger.trace("chord")
        if self.board[clue_pos] in [VALUE, MINES, self.clue_tag, self.flag_tag, None]:
            return []
        if self.board == self.last_hint[0]:
            for objs, positions in self.last_hint[1].items():
                if len(objs) == 1 and objs[0] == clue_pos:
                    return positions
            return []
        obj = self.board.get_value(clue_pos)
        board: AbstractBoard = self.board.clone()
        chord_positions = []
        if obj.deduce_cells(board) is not None:
            for pos, obj in self.board():
                if (obj is None) and (board[pos] is not None):
                    chord_positions.append(pos)
            return chord_positions

        for pos, obj_type in board(mode="type"):
            if clue_pos == pos:
                continue
            if obj_type == "C":
                board[pos] = self.clue_tag
            elif obj_type == "F":
                board[pos] = MINES_TAG

        chord_positions = self._deduced(board, [
            self.summon.mines_clue_rule
            if self.board.get_type(clue_pos) == "F" else
            self.summon.clue_rule,
        ])

        self.logger.trace(f"chord pos: {clue_pos}, {self.board[clue_pos]}, {chord_positions}")

        return chord_positions

    def apply(self, pos: AbstractPosition, action: int) -> Union["AbstractBoard", None]:
        """
        :param pos: 交互位置
        :param action: 操作代码
            0: 左键点击/翻开/设置非雷
            1: 右键点击/标雷/设置必雷
        """
        global NORMAL, EXPERT, ULTIMATE, PUZZLE
        _board = None
        if self.mode == PUZZLE:
            if action == 1:
                value_tag = self.board.get_config(pos.board_key, "MINES")
                self.board.set_value(pos, self.flag_tag if value_tag == VALUE_QUESS else value_tag)
            elif action == 0:
                value_tag = self.board.get_config(pos.board_key, "VALUE")
                self.board.set_value(pos, self.clue_tag if value_tag == VALUE_QUESS else value_tag)
            return self.board
        if self.board.get_type(pos) != "N":
            # 点击了线索
            chord_positions = self.chord_clue(pos)
            if self.mode in [NORMAL, EXPERT]:
                # 普通和专家直接设置值
                for _pos in chord_positions:
                    self.board[_pos] = self.answer_board[_pos]
            elif self.mode in [ULTIMATE, PUZZLE]:
                # 如果是纸笔和专家就放标志
                for _pos in chord_positions:
                    if _pos.board_key not in self.board.get_interactive_keys():
                        if self.answer_board.get_type(_pos) == "C":
                            self.board[_pos] = self.board.get_config(_pos.board_key, "VALUE")
                        elif self.answer_board.get_type(_pos) == "F":
                            self.board[_pos] = self.board.get_config(_pos.board_key, "MINES")
                    elif self.answer_board.get_type(_pos) == "F":
                        self.board[_pos] = self.flag_tag
                    elif self.answer_board.get_type(_pos) == "C":
                        self.board[_pos] = self.clue_tag

        elif self.mode == NORMAL:
            # 普通模式
            if not action and self.answer_board.get_type(pos) == "F":
                return None
            if action and self.answer_board.get_type(pos) == "C":
                return None
            self.board[pos] = self.answer_board[pos]
        elif self.mode in [EXPERT, ULTIMATE, PUZZLE]:
            # 专家模式
            _board = self.board.clone()
            if pos not in self.last_deduced[1]:
                print(f"apply {pos} 未命中 {self.last_deduced[1]}")
                if pos not in self.deduced():
                    return None
            if action and self.answer_board.get_type(pos) == "C":
                return None
            if not action and self.answer_board.get_type(pos) == "F":
                return None
            if self.mode in [ULTIMATE, PUZZLE]:
                self.board[pos] = self.flag_tag if action else self.clue_tag
            else:
                self.board[pos] = self.answer_board[pos]
        else:
            return None
        if pos.board_key not in self.board.get_interactive_keys():
            if self.board[pos] is self.clue_tag:
                self.board[pos] = self.board.get_config(pos.board_key, "VALUE")
            elif self.board[pos] is self.flag_tag:
                self.board[pos] = self.board.get_config(pos.board_key, "MINES")
        if (
            self.mode in [ULTIMATE] and
            self.ultimate_mode & ULTIMATE_A
        ):
            self.step()
            if self.drop_r:
                flag = True
                if self.ultimate_mode & ULTIMATE_F:
                    for pos in self.deduced():
                        if self.answer_board.get_type(pos) != "F":
                            continue
                        if pos.board_key not in self.board.get_interactive_keys():
                            continue
                        flag = False
                        break
                if self.ultimate_mode & ULTIMATE_S:
                    for pos in self.deduced():
                        if pos.board_key in self.board.get_interactive_keys():
                            continue
                        flag = False
                        break
                for pos in self.deduced():
                    if pos.board_key not in self.board.get_interactive_keys():
                        continue
                    if self.answer_board.get_type(pos) == "F":
                        continue
                    flag = False
                    break
                if flag:
                    self.drop_r = False
                    self.last_deduced[0] = None
                    self.deduced()
            self.thread_deduced()
        return self.board

    def click(self, pos: "AbstractPosition") -> Union["AbstractBoard", None]:
        """
        翻开/点击 某个空白格
        :param pos: 翻开的位置
        """
        return self.apply(pos, 0)

    def mark(self, pos: AbstractPosition) -> Union["AbstractBoard", None]:
        """
        右键标雷
        """
        return self.apply(pos, 1)

    def step(self) -> bool:
        print("step")
        if self.ultimate_mode & ULTIMATE_F:
            for pos in self.deduced():
                if self.answer_board.get_type(pos) != "F":
                    continue
                if pos.board_key not in self.board.get_interactive_keys():
                    continue
                print("[step]has flag")
                return False
        if self.ultimate_mode & ULTIMATE_S:
            for pos in self.deduced():
                if pos.board_key in self.board.get_interactive_keys():
                    continue
                print("[step]has mark")
                return False
        for pos in self.deduced():
            if pos.board_key not in self.board.get_interactive_keys():
                continue
            if self.answer_board.get_type(pos) == "F":
                continue
            print("[step]has clue")
            return False
        for key in self.board.get_board_keys():
            for pos, obj in self.board(key=key):
                if obj not in [self.clue_tag, self.flag_tag]:
                    continue
                self.board[pos] = self.answer_board[pos]

        self.last_deduced[0] = None
        self.last_hint[0] = None
        self.thread_hint()
        return True

    def deduced(self, thread=True):
        """
        收集所有必然能推出的位置及其不可能的值
        """
        if self.last_deduced[0] == self.board:
            return self.last_deduced[1]
        # 如果是终极模式
        if self.mode == ULTIMATE and self.last_deduced[1]:
            for pos in self.last_deduced[1][:]:
                if self.board[pos] is not None:
                    self.last_deduced[1].remove(pos)
            if self.ultimate_mode & ULTIMATE_F:
                for pos in self.last_deduced[1]:
                    if self.answer_board.get_type(pos) != "F":
                        continue
                    if pos.board_key not in self.board.get_interactive_keys():
                        continue
                    print("[deduced]has flag")
                    return self.last_deduced[1]
            if self.ultimate_mode & ULTIMATE_S:
                for pos in self.last_deduced[1]:
                    if pos.board_key in self.board.get_interactive_keys():
                        continue
                    print("[deduced]has mark")
                    return self.last_deduced[1]
            for pos in self.last_deduced[1]:
                if pos.board_key not in self.board.get_interactive_keys():
                    continue
                if self.answer_board.get_type(pos) == "F":
                    continue
                print("[deduced]has clue")
                return self.last_deduced[1]
        self.deduced_queue.put("process")  # 请求处理权
        try:
            t = time.time()
            self.logger.debug(f"deduced start (last: {str(self.last_deduced[1])})")
            deduced = []
            board = self.board.clone()
            for pos in self.last_deduced[1]:
                if board.get_type(pos) != "N":
                    continue
                if self.answer_board.get_type(pos) == "F":
                    board[pos] = self.flag_tag
                elif self.answer_board.get_type(pos) == "C":
                    board[pos] = self.clue_tag
                deduced.append(pos)
            all_rules = self.summon.mines_rules.rules[:]
            all_rules += [self.summon.clue_rule, self.summon.mines_clue_rule]
            deduced += self._deduced(board, all_rules)

            self.last_deduced[1] = deduced
            self.last_deduced[0] = self.board.clone()

            self.logger.debug(f"last_deduced {str(self.last_deduced[1])}")
            self.logger.debug(f"deduced used time {time.time() - t}")

            return deduced
        finally:
            self.deduced_queue.get()
            self.deduced_queue.task_done()

    def hint(self, thread=False):
        if self.last_hint[0] == self.board:
            return self.last_hint[1]

        if None not in [obj for pos, obj in self.board()]:
            return {}

        deduced = self.deduced()
        if not deduced and thread:
            return {}
        if not deduced and self.mode != ULTIMATE:
            self.logger.error("题板无可推格")
            return {}
        if self.mode == ULTIMATE and not deduced:
            _board = self.board.clone()
            self.step()
            positions = []
            if not (self.drop_r and self.deduced()):
                self.drop_r = False
                positions.append(("R", None))
            self.last_deduced[0] = None
            self.last_hint[0] = None
            for pos, obj in _board():
                if type(obj) in [type(self.clue_tag), type(self.flag_tag)]:
                    positions.append(pos)
            print("step: ", positions)
            return {tuple(positions): []}

        self.hint_queue.put("process")  # 请求处理权
        try:
            board = self.board.clone()
            t = time.time()
            hint = self._hint(board)
            self.logger.debug(f"all_hint {hint}")
            self.logger.debug(f"hint used_time {time.time() - t}")

            return hint
        finally:
            self.hint_queue.get()
            self.hint_queue.task_done()

    def _deduced(self, board, all_rules):
        self.logger.trace("构建新模型")
        self.logger.trace(f"deduced all_rules: {all_rules}")
        self.logger.trace(f"deduced drop_r: {self.drop_r}")
        board.clear_variable()
        model = board.get_model()
        switch = Switch()

        # 2.获取所有规则约束
        for rule in all_rules:
            if rule is None:
                continue
            if self.drop_r and isinstance(rule, Rule0R):
                continue
            rule.create_constraints(board, switch)

        for key in board.get_board_keys():
            for pos, obj in board(key=key):
                if obj is None:
                    continue
                obj.create_constraints(board, switch)

        # 3.获取所有变量并赋值已解完的部分
        for key in board.get_board_keys():
            for _, var in board("C", mode="variable", key=key):
                model.Add(var == 0)
                self.logger.trace(f"var: {var} == 0")
            for _, var in board("F", mode="variable", key=key):
                model.Add(var == 1)
                self.logger.trace(f"var: {var} == 1")

        for switch_var in switch.get_all_vars():
            model.Add(switch_var == 1)

        results = []
        future_to_param = {}

        with ThreadPoolExecutor(max_workers=CONFIG["workes_number"]) as executor:
            # 提交任务
            for key in board.get_board_keys():
                for pos, _ in board("N", key=key):
                    fut = executor.submit(
                        deduced_by_csp,
                        board,
                        self.answer_board,
                        pos
                    )
                    future_to_param[fut] = pos  # 记录参数以便出错追踪

            # 收集结果
            for fut in as_completed(future_to_param):
                pos = future_to_param[fut]
                try:
                    self.logger.trace(f"deduced pos {pos} wait")
                    result = fut.result()
                    self.logger.trace(f"deduced pos {pos} end: {result}")
                    if result is None:
                        continue
                    if result:
                        results.append(pos)
                except Exception as exc:
                    raise exc

        self.logger.trace(f"deduced done: {results}")

        return results

    def _hint(self, board) -> dict[tuple, list[AbstractPosition]]:
        """
        返回每一类推理依据及其能推出的位置
        """
        deduced = self.deduced()
        get_random().shuffle(deduced)
        future_to_param = {}
        results = {}
        with ThreadPoolExecutor(max_workers=CONFIG["workes_number"]) as executor:
            # 提交任务
            for pos, obj in board("CF"):
                fut = executor.submit(
                    self.chord_clue,
                    pos
                )
                future_to_param[fut] = pos  # 记录参数以便出错追踪

            # 收集结果
            for fut in as_completed(future_to_param):
                pos = future_to_param[fut]
                try:
                    result = fut.result()
                    if not result:
                        continue
                    results[tuple([pos])] = result
                except Exception as exc:
                    raise exc
        if results:
            self.last_hint[1] = results
            self.last_hint[0] = board.clone()
            return results

        self.logger.trace("构建新模型")
        board.clear_variable()
        model = board.get_model()
        switch = Switch()

        all_rules = self.summon.mines_rules.rules[:]
        all_rules += [self.summon.clue_rule, self.summon.mines_clue_rule]

        # 2.获取所有规则约束
        for rule in all_rules:
            if rule is None:
                continue
            if self.drop_r and isinstance(rule, Rule0R):
                continue
            rule.create_constraints(board, switch)

        for key in board.get_board_keys():
            for pos, obj in board(key=key):
                if obj is None:
                    continue
                if obj.invalid(board):
                    continue
                obj.create_constraints(board, switch)

        # 3.获取所有变量并赋值已解完的部分
        for key in board.get_board_keys():
            for _, var in board("C", mode="variable", key=key):
                model.Add(var == 0)
                self.logger.trace(f"var: {var} == 0")
            for _, var in board("F", mode="variable", key=key):
                model.Add(var == 1)
                self.logger.trace(f"var: {var} == 1")

        results = {}
        future_to_param = {}
        # upper_bound = None
        upper_bound = [float("inf"), threading.Lock()]

        with ThreadPoolExecutor(max_workers=CONFIG["workes_number"]) as executor:
            # 提交任务
            for pos in deduced:
                fut = executor.submit(
                    hint_by_csp, board,
                    self.answer_board,
                    switch, pos, upper_bound
                )
                future_to_param[fut] = pos  # 记录参数以便出错追踪

            # 收集结果
            for fut in as_completed(future_to_param):
                pos = future_to_param[fut]
                try:
                    self.logger.trace(f"pos[{pos}]: start")
                    _result = fut.result()
                    self.logger.trace(f"pos[{pos}]: {_result}")
                    if _result is None:
                        continue
                    self.logger.trace(deduced)
                    _result.sort()
                    _result = list(tuple(set(_result)))
                    result = set()
                    for k in _result:
                        bes_type = k[0].split("|", 1)[0]
                        name = k[0].split("|", 1)[1]
                        if bes_type == "RULE":
                            result.add((name, k[1]))
                        elif bes_type == "POS":
                            info = name.split("|", 2)
                            result.add(
                                board.get_pos(
                                    int(info[0]),
                                    int(info[1]),
                                    info[2]
                                ))
                    result = tuple(result)
                    if result not in results:
                        results[result] = []
                    results[result].append(pos)
                except Exception as exc:
                    raise exc

        self.last_hint[0] = board.clone()
        self.last_hint[1] = results

        return results

    def check_difficulty(self, q=1000, br=False):
        clue_freq = {1: 0}
        _board = self.board.clone()
        n_num = len([None for key in _board.get_board_keys()
                     for _ in _board('N', key=key)])
        while self.board.has("N"):
            if br and max(clue_freq.keys()) >= q:
                return clue_freq

            print(f"{n_num - len([None for key in self.board.get_board_keys() for _ in self.board('N', key=key)])}"
                  f"/{n_num}", end="\r")
            num_clues_used = float("inf")

            n_length = len([None for key in self.board.get_board_keys() for _ in self.board('N', key=key)])
            print(f"{n_num - n_length}/{n_num}", end="\r")
            self.logger.debug("\n" + self.board.show_board())
            self.logger.debug(clue_freq)
            grouped_hints = self.hint().items()
            if not grouped_hints:
                self.logger.warn("hint无返回值")
                self.logger.warn("\n" + self.board.show_board())
                return
            self.logger.debug("\n" + self.board.show_board())
            [self.logger.debug(str(i[0]) + " -> " + str(i[1])) for i in grouped_hints]
            pos_clues = {}
            for hints, deduceds in grouped_hints:
                if "R" in hints:
                    hints_length = 1 + (len(hints) // 4)
                else:
                    hints_length = len(hints)
                if hints_length > num_clues_used:
                    continue
                elif hints_length < num_clues_used:
                    num_clues_used = hints_length
                    pos_clues.clear()
                for deduced in deduceds:
                    pos_clues[deduced] = num_clues_used
            for pos in pos_clues:
                imposs = self.answer_board.get_type(pos)
                self.apply(pos, 0 if imposs == "C" else 1)
                if pos_clues[pos] not in clue_freq:
                    clue_freq[pos_clues[pos]] = 0
                clue_freq[pos_clues[pos]] += 1
            self.logger.debug("\n" + self.board.show_board())
            self.logger.debug(clue_freq)
        self.board = _board
        return clue_freq


def main():
    get_logger(log_lv="TRACE")
    # get_random(new=True, seed=8894987)
    # get_random(seed=5474554)
    size = (5, 5)
    rules = ["*3T"]
    s = Summon(size, -1, rules)
    g = GameSession(s, ULTIMATE, False, ULTIMATE_R)
    # g.board = s.create_puzzle()
    # g.answer_board = s.answer_board
    g.answer_board = s.summon_board()
    g.create_board()
    # for p, i in g.board("C"):
    #     g.chord_clue(p)
    # g.drop_r = True
    print("=" * 20)
    print(g.board)
    print("=" * 20)
    print(d := g.deduced())
    print("=" * 20)
    print(g.board)
    print("=" * 20)
    # print(f"deduced: {d}")
    print(h := g.hint())
    print("=" * 20)
    print(g.board)
    print("=" * 20)
    for b, t in h.items():
        print(b, "->", t)
    # g.create_board()
    # while "N" in g.board:
    #     print(g.hint())
    #     print(g.deduced_values)
    #     print(g.board)
    #     for p, v in list(g.deduced_values.items())[:]:
    #         if v is MINES_TAG:
    #             g.click(p)
    #         else:
    #             g.mark(p)
    # print(g.__dict__)
    # print(g.answer_board)
    # print(g.answer_board[g.answer_board.get_pos(1, 1)].high_light(g.answer_board))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:42
# @Author  : Wu_RH
# @FileName: 2C.py
"""
[2C] 连方 (Connected)：(1) 所有四连通雷区域为矩形；(2) 所有雷区域对角相邻
"""
from typing import List

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition
from .connect import connect


def block(a_pos: AbstractPosition, board: AbstractBoard) -> List[AbstractPosition]:
    b_pos = a_pos.up()
    c_pos = a_pos.left()
    d_pos = b_pos.left()
    if not board.in_bounds(d_pos):
        return []
    return [a_pos, b_pos, c_pos, d_pos]


class Rule2C(AbstractMinesRule):
    name = ["2C", "连方", "Connected"]
    doc = "(1) 所有四连通雷区域为矩形；(2) 所有雷区域对角相邻"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s1 = switch.get(model, self)
        s2 = switch.get(model, self)
        for pos, _ in board():
            var_list = board.batch(block(pos, board), mode="variable")
            if not var_list:
                continue
            model.Add(sum(var_list) != 3).OnlyEnforceIf(s1)
        connect(
            model,
            board,
            s2
        )



import time
import uuid
from orjson import loads, dumps
from typing import Awaitable

from flask.typing import ResponseReturnValue, RouteCallable
from flask import make_response, request
from .model import Model
from .datastore import DataStore

class SessionManager:
    def __init__(self, db: DataStore, model: type[Model]):
        self.db = db
        self.data = {}
        self.model = model

        self.host: Model| None = None

    def get(self, token: str):
        if token in self.data:
            return self.data[token]
        return None

    @staticmethod
    def gen_token():
        return str(uuid.uuid4()), time.time()

    async def new_token(self):
        token, created_at = self.gen_token()
        info = {"created_at": created_at}
        await self.db.set(token, info)
        self.data[token] = {"info": info}
        return token

    async def create(self):
        token = await self.new_token()

        data = self.get(token)

        if data is None:
            raise RuntimeError("Session data not found")

        if self.host is None:
            data["game"] = self.host = self.model(token=token)
        else:
            data["game"] = self.model(host=self.host, token=token)

        return token, data

    def wrapper(self, func: RouteCallable) -> RouteCallable:
        async def _func() -> ResponseReturnValue:
            token = request.args.get("token")

            if token is None or not (data := self.get(token)):
                return ('Unauthorized', 401)

            result = func(data["game"])

            if isinstance(result, Awaitable):
                result = await result

            return result
        return _func

    def gen_token_route(self) -> RouteCallable:
        async def _func() -> ResponseReturnValue:
            token, _ = await self.create()
            return {"token": token, "success": True}
        return _func



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/20 14:12
# @Author  : Wu_RH
# @FileName: 2#.py
"""
[2#]: 包含以下规则:[V], [2X], [2X'], [2D], [2P], [2M], [2A]
注: 通过"2#:"来去除2A
"""
from minesweepervariants.abs.board import AbstractBoard
from . import AbstractClueSharp


class Rule2sharp(AbstractClueSharp):
    name = ["2#'", "标签"]
    doc = "包含以下规则:[V], [2X], [2X'], [2D], [2P], [2M], [2A]"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        rules_name = ["V", "2X", "2X'", "2D", "2P", "2M"]
        if data is None:
            rules_name += ["2A"]
        super().__init__(rules_name, board, data)





#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/09 11:37
# @Author  : xxx
# @FileName: 3A.py
"""
[3A]兰顿蚂蚁: 数字表示兰顿蚂蚁从线索格出发直至走出题板外所移动的次数。
箭头表示兰顿蚂蚁的初始方向，经过非雷格顺时针旋转90度(右转)，经过雷格逆时针旋转90度(左转)。
数字可以为无穷,其表示当前路径构成了循环
"""
from typing import List, Dict

from minesweepervariants.utils.web_template import StrWithArrow

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.image_create import get_image, get_text, get_row, get_col, get_dummy
from ....utils.impl_obj import MINES_TAG, VALUE_QUESS
from ....utils.tool import get_random


def put(board, pos: 'AbstractPosition', path):
    value = 0
    while board.in_bounds(pos):
        value += 1
        if board.get_type(pos) == "F":
            path += 1
            path %= 4
        else:
            obj = board.get_value(pos)
            obj.append((value, path))
            path += 3
            path %= 4
        # 0.上 1.右 2.下 3.左
        match path:
            case 0:
                pos = pos.down()
            case 1:
                pos = pos.left()
            case 2:
                pos = pos.up()
            case 3:
                pos = pos.right()


class Rule3A(AbstractClueRule):
    name = ["3A", "兰顿蚂蚁"]
    doc = "数字表示兰顿蚂蚁从线索格出发直至走出题板外所移动的次数。"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        for pos, _ in board("N"):
            board.set_value(pos, Value3A(pos))
        lines = [
            board.get_row_pos(board.get_pos(0, 0)),
            board.get_col_pos(board.get_pos(-1, -1)),
            board.get_row_pos(board.get_pos(-1, -1)),
            board.get_col_pos(board.get_pos(0, 0)),
        ]
        for index in range(4):
            line = lines[index]
            for pos in line:
                put(board, pos, index)
                # print(pos, index)
                # print(board.show_board())
        for _, obj in board("C"):
            obj.end(random.random())
        return board


class Value3A(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        if not code:
            self.value = 0
            self.dir = 0
            self.data = []
            return
        self.value = code[1]  # 路径长
        self.dir = code[0]  # 0.上 1.右 2.下 3.左

    def __str__(self) -> str:
        return f"{self.value}{'↑→↓←'[self.dir]}"

    def __repr__(self) -> str:
        return f"{self.value}{'^>v<'[self.dir]}"

    def web_component(self, board) -> Dict:
        return StrWithArrow(str(self.value), ["up", "right", "down", "left"][self.dir])

    def compose(self, board) -> Dict:
        match self.dir:
            case 0:
                # 上 ↑ ^
                if self.value == 1:
                    return get_row(
                            get_dummy(width=0.175),
                            get_text("1"),
                            get_image("up"),
                            get_dummy(width=0.175),
                        )
                return get_row(
                        get_text(str(self.value)),
                        get_image("up"),
                        spacing=-0.1,
                    )
            case 1:
                # 右 → >
                return get_col(
                    get_dummy(height=0.1),
                    get_image(
                        "right",
                        image_height=0.2,
                        image_width=0.7
                    ),
                    get_dummy(height=-0.05),
                    get_text(str(self.value)),
                    get_dummy(height=0.2)
                )
            case 2:
                # 下 ↓ V
                if self.value == 1:
                    return get_row(
                            get_dummy(width=0.175),
                            get_text("1"),
                            get_image("down"),
                            get_dummy(width=0.175),
                        )
                return get_row(
                        get_text(str(self.value)),
                        get_image("down"),
                        spacing=-0.1,
                    )
            case 3:
                # 左 ← <
                return get_col(
                    get_dummy(height=0.1),
                    get_image(
                        "left",
                        image_height=0.2,
                        image_width=0.7
                    ),
                    get_dummy(height=-0.05),
                    get_text(str(self.value)),
                    get_dummy(height=0.2)
                )
        return get_text("")

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition']:
        pos = self.pos.clone()
        path = (self.dir + 1) % 4
        position = []
        value = float("inf") if self.value == 0 else self.value
        while True:
            if not board.in_bounds(pos):
                break
            if board.get_type(pos) == "N":
                position.append(pos)
                break
            position.append(pos)
            # 0.上 1.右 2.下 3.左
            if board.get_type(pos) == "F":
                path += 3
                path %= 4
            else:
                path += 1
                path %= 4
            match path:
                case 0:
                    pos = pos.down()
                case 1:
                    pos = pos.left()
                case 2:
                    pos = pos.up()
                case 3:
                    pos = pos.right()
            if (
                pos == self.pos and
                path == (self.dir + 1) % 4
            ):
                break
            value -= 1
            if value == 0:
                break
        # print(self.value, position)
        return position

    def append(self, data):
        self.data.append(data)

    def end(self, rand: float):
        if len(self.data) == 0:
            self.dir = int(rand * 4)
            return

        self.data.sort(key=lambda x: x[0])
        self.data: list[tuple[int, int]]

        # 提取所有首个int值
        values = [item[0] for item in self.data]

        # 计算权重 (n_i^2) 和总权重
        weights = [n ** 2 for n in values]
        total_weight = sum(weights)

        # 构建累积概率分布
        cumulative = 0.0
        for i, weight in enumerate(weights):
            prob = weight / total_weight  # 当前元素的概率
            cumulative += prob  # 累积概率
            # 如果 rand 小于当前累积概率，选择当前元素
            if rand < cumulative:
                self.value, self.dir = self.data[i]
                self.data = None
                break

        if self.data is not None:
            self.value, self.dir = self.data[-1]
        del self.data

    @classmethod
    def type(cls) -> bytes:
        return Rule3A.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.dir, self.value])

    def bfs_get_states(self, board: 'AbstractBoard') -> List[dict]:
        """
        大致思路:
        从方向开始走 如果提前走出边境或者循环就直接丢弃 如果刚好在值内就记录

        一层具象
        从当前位置为根节点构建一棵二叉树进行遍历
        如果说value==-1 那么需要进入while遍历并把所有走出题板的节点删除
        如果value>=1 那么就使用range()

        二层具象
        如果值为无穷 那么需要使用快慢节点进行遍历 一旦重合就立刻将其值放入list
        可能会遇到重复的pos节点 在遍历的时候需要边查表边遍历
        如果出现过就需要直接赋值而不是继续分叉

        三层具象
        针对>=1可以使用ab_bfs列表 对当前abfs进行遍历
        如果出现超出题板 那么就丢出
        如果出现循环 不用管 等时间到了他就死了
        每次创建叉的时候需要查看当前经过的路径是否遍历过

        >=1:
            需要附带的信息有 当前节点, 当前方向, 遍历过的路径的对应值
            具体步骤为:
                沿着当前方向走一步
                查看当前节点有没有超出题板内
                    超出题板就撇了
                查看当前节点在不在对应值列表里面
                    不在就加入对应值列表
                加入b_bfs
        """

        def move(_pos, _dir):
            match _dir:
                case 0:
                    _pos = _pos.up()
                case 1:
                    _pos = _pos.right()
                case 2:
                    _pos = _pos.down()
                case 3:
                    _pos = _pos.left()
            return _pos

        def node(_pos, _dt, _value_map):
            result = []
            if _pos in _value_map:
                # 如果在就继续走
                result.append((_pos, (_dt + _value_map[_pos] * 2 + 1) % 4, _value_map.copy()))
            else:
                if (pos_type := board.get_type(_pos)) != "N":
                    # 当前位置是有值的 直接拿了用
                    _value_map = _value_map.copy()
                    _value_map[_pos] = 0 if pos_type == "C" else 1
                    result.append((_pos, (_dt + _value_map[_pos] * 2 + 1) % 4, _value_map))
                else:
                    # 如果不在说明未经过过该节点 就加入map
                    _value_map_a = _value_map.copy()
                    _value_map_b = _value_map.copy()
                    _value_map_a[_pos] = 1
                    _value_map_b[_pos] = 0
                    result.extend([
                        (_pos, (_dt + 3) % 4, _value_map_a),
                        (_pos, (_dt + 1) % 4, _value_map_b),
                    ])
            return result

        root = self.pos.clone()
        a_bfs = [(root, self.dir, {root.clone(): 0})]
        b_bfs = []
        answer_list = []

        if self.value == 0:
            flag = False
            # 处理循环的情况
            while len(a_bfs) > 0:
                for pos, dt, value_map in a_bfs:
                    if flag and pos == self.pos and dt == self.dir:
                        # 如果当前位置和方向和自身完全相同 就说明循环了
                        answer_list.append(value_map)
                        continue
                    # 沿着当前方向走一步
                    # 0.上 1.右 2.下 3.左
                    pos = move(pos, dt)
                    if not board.in_bounds(pos):
                        # 查看当前节点有没有超出题板内
                        continue
                    b_bfs.extend(node(pos, dt, value_map))
                a_bfs = b_bfs
                b_bfs = []
                flag = True
            return answer_list
        else:
            for depth in range(self.value, 0, -1):
                for pos, dt, value_map in a_bfs:
                    # 沿着当前方向走一步
                    # 0.上 1.右 2.下 3.左
                    pos = move(pos, dt)
                    # 查看当前节点有没有超出题板内
                    if not board.in_bounds(pos):
                        # 如果depth==self.value就是要的这个结果
                        if depth == 1:
                            answer_list.append(value_map)
                        continue
                    b_bfs.extend(node(pos, dt, value_map))
                a_bfs = b_bfs
                b_bfs = []
            return answer_list

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        # 开玩笑 还真能写
        answer_list = self.bfs_get_states(board)
        pos_map = {}
        for line in answer_list:
            for pos in line:
                if pos not in pos_map:
                    pos_map[pos] = [line[pos], 1]
                else:
                    if pos_map[pos][0] == line[pos]:
                        pos_map[pos][1] += 1
        change = False
        for pos in pos_map:
            if pos_map[pos][1] == len(answer_list):
                if board.get_type(pos) == "N":
                    if pos_map[pos][0]:
                        board.set_value(pos, MINES_TAG)
                    else:
                        board.set_value(pos, VALUE_QUESS)
                    change = True
        return change

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        answer_list = self.bfs_get_states(board)
        var_list = []
        for index in range(len(answer_list)):
            line = answer_list[index]
            var = model.NewBoolVar(f"[3A]{self.pos}{index}")
            for pos, value in line.items():
                model.Add(board.get_variable(pos) == value).OnlyEnforceIf([var, s])
            var_list.append(var)
        if len(var_list) == 0:
            # 无解
            var = model.NewBoolVar(f"[3A]Error")
            model.Add(var == 1).OnlyEnforceIf(s)
            model.Add(var == 0).OnlyEnforceIf(s)
        model.Add(sum(var_list) == 1).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/11 16:38
# @Author  : Wu_RH
# @FileName: 2E'.py
"""
[1K2E']自指:如果字母X马步8格内有N个雷，则标有X=N的格子必定是雷。
"""
from .....utils.impl_obj import VALUE_QUESS
from .....utils.tool import get_random, get_logger

from .....abs.Rrule import AbstractClueValue, AbstractClueRule
from .....abs.board import AbstractBoard, AbstractPosition, MASTER_BOARD

ALPHABET = "ABCDEFGHI"


class Rule1K2Ep(AbstractClueRule):
    name = ["1K2E'", "马步自指"]
    doc = "如果字母X马步8格内有N个雷，则标有X=N的格子必定是雷。"

    def __init__(self, board: AbstractBoard, data=None):
        super().__init__()
        board.set_config(MASTER_BOARD, "pos_label", True)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        letter_map = {i: [] for i in range(9)}
        logger = get_logger()
        for pos, _ in board("F"):
            if pos.y > 8:
                continue
            letter = ALPHABET[pos.y]
            if pos.x not in letter_map:
                letter_map[pos.x] = []
            letter_map[pos.x].append(letter)

        for pos, _ in board("N"):
            positions = pos.neighbors(5, 5)
            value = board.batch(positions, mode="type", drop_none=True).count("F")
            if not letter_map[value]:
                board.set_value(pos, VALUE_QUESS)
                continue
            letter = random.choice(letter_map[value])
            obj = Value1K2Ep(pos, bytes([ALPHABET.index(letter)]))
            board.set_value(pos, obj)
            logger.debug(f"[2E'] put {letter}({value}) at {pos}")
        return board


class Value1K2Ep(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        self.value = code[0]    # 实际为第几列的字母
        self.neighbors = pos.neighbors(5, 5)
        self.pos = pos

    def __repr__(self) -> str:
        return f"{ALPHABET[self.value]}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbors

    @classmethod
    def type(cls) -> bytes:
        return Rule1K2Ep.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        pos = board.get_pos(0, self.value)
        line = board.get_col_pos(pos)
        line = board.batch(line, mode="variable")
        sum_vers = sum(board.batch(self.neighbors, mode="variable", drop_none=True))
        for index in range(min(9, len(line))):
            var = board.get_variable(board.get_pos(index, self.value))
            model.Add(sum_vers != index).OnlyEnforceIf(var.Not()).OnlyEnforceIf(s)



"""
[SC] 甘蔗 (Sugar Cane)：每个雷周围四格至少有一个染色格, 且染色格不能为雷
"""

from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch

class RuleSC(AbstractMinesRule):
    name = ["SC", "甘蔗", "Sugar Cane"]
    doc = "每个雷周围至少有一个染色格, 且染色格不能为雷"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="var"):
            if board.get_dyed(pos):
                model.Add(var == 0).OnlyEnforceIf(s)
                continue
            var_list = board.batch(pos.neighbors(1), mode="dye", drop_none=True)
            if not any(var_list):
                model.Add(var == 0).OnlyEnforceIf(s)


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/07 22:21
# @Author  : Wu_RH
# @FileName: UP.py
"""
[UP]唯一路径(Unique Path): 线索格表示从这个格开始只能往右或下走，到达右下角的方法数。
"""
import math
from typing import List

from minesweepervariants.abs.Rrule import AbstractClueRule, AbstractClueValue
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch


class RuleUP(AbstractClueRule):
    name = ["UP", "唯一路径", "Unique Path"]
    doc = "线索格表示从这个格开始只能往右或下走，到达右下角的方法数。"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for key in board.get_interactive_keys():
            root_pos = board.boundary(key)
            for col_pos in board.get_col_pos(root_pos)[::-1]:
                for pos in board.get_row_pos(col_pos)[::-1]:
                    if board.get_type(pos) != "N":
                        continue
                    value = 0
                    if board.get_type(pos.down()) == "C":
                        value += board[pos.down()].value
                    if board.get_type(pos.right()) == "C":
                        value += board[pos.right()].value
                    if (
                        board.get_type(pos.down()) == "" and
                        board.get_type(pos.right()) == ""
                    ):
                        value = 1
                    obj = ValueUP(pos, bytes([value]))
                    board[pos] = obj
        return board

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        pos_var_map = {}
        for key in board.get_interactive_keys():
            ub_x = board.boundary(key).x
            ub_y = board.boundary(key).y
            for pos, var in board(mode="variable", key=key):
                x, y = pos.x, pos.y
                ub = math.comb(ub_x - x + ub_y - y, ub_y - y)
                pos_var_map[pos] = model.NewIntVar(0, ub, f"{pos}:dp")
                model.Add(pos_var_map[pos] == 0).OnlyEnforceIf(var)

            for pos, var in board(mode="variable", key=key):
                var_d = 0
                var_r = 0
                if board.in_bounds(pos.down()):
                    var_d = pos_var_map[pos.down()]
                if board.in_bounds(pos.right()):
                    var_r = pos_var_map[pos.right()]
                if board.boundary(key) == pos:
                    model.Add(pos_var_map[pos] == 1).OnlyEnforceIf(var.Not())
                else:
                    model.Add(pos_var_map[pos] == var_d + var_r).OnlyEnforceIf(var.Not())
                obj = board[pos]
                if isinstance(obj, ValueUP):
                    obj.create_constraints_(board, pos_var_map[pos], switch)


class ValueUP(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.pos = pos
        self.value = code[0]

    def __repr__(self):
        return str(self.value)

    def high_light(self, board: 'AbstractBoard', pos=None) -> List['AbstractPosition'] | None:
        if pos is None:
            pos = self.pos
        if board.get_type(pos) != "C":
            return [pos]
        positions = {pos}
        for _pos in self.high_light(board, pos.down()):
            positions.add(_pos)
        for _pos in self.high_light(board, pos.right()):
            positions.add(_pos)
        return list(positions)

    @classmethod
    def type(cls) -> bytes:
        return RuleUP.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints_(self, board: 'AbstractBoard', var, switch: 'Switch'):
        model = board.get_model()
        model.Add(var == self.value).OnlyEnforceIf(switch.get(model, self))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2025/06/03 01:58
# @Author  : Wu_RH
# @FileName: board.py

from typing import List, Union, Tuple, Any, Generator, TYPE_CHECKING
import heapq

import gc
from ortools.sat.python import cp_model
from ortools.sat.python.cp_model import IntVar

from ....abs.rule import AbstractValue
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG
from ....utils.impl_obj import POSITION_TAG, VALUE_CROSS, VALUE_CIRCLE
from ....utils.tool import get_logger
from ....abs.board import AbstractBoard, AbstractPosition, MASTER_BOARD
from ....abs.Rrule import AbstractClueValue
from ....abs.Mrule import AbstractMinesValue

if TYPE_CHECKING:
    ...


def get_value(pos=None, code=None):
    from minesweepervariants.impl.impl_obj import get_value
    return get_value(pos, code)


def alpha(n: int) -> str:
    alpha_map = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    if n < 26:
        return alpha_map[n]
    return alpha_map[n // 26 - 1] + alpha_map[n % 26]


def encode_int_7bit(n: int) -> bytes:
    # 编码主体：每7位 -> 1字节（bit6~bit0，bit7=0）
    if n == 0:
        return b'\x00'
    payload = []

    while n > 0:
        payload.append(n & 0x7f)
        n >>= 7

    return bytes(payload)


def decode_bytes_7bit(data: bytes) -> int:
    if len(data) == 0:
        return 0

    result = 0
    for i in data[::-1]:
        result <<= 7
        result |= i

    return result


class Position(AbstractPosition):
    def __repr__(self):
        return (f"{self.board_key+':' if self.board_key != MASTER_BOARD else ''}"
                f"{alpha(self.y)}{self.x+1}")

    def _up(self, n: int = 1):
        self.x -= n

    def _down(self, n: int = 1):
        self.x += n

    def _left(self, n: int = 1):
        self.y -= n

    def _right(self, n: int = 1):
        self.y += n

    def _deviation(self, pos: 'Position'):
        self.x += pos.x
        self.y += pos.y

    def in_bounds(self, bound_pos: 'Position') -> bool:
        if bound_pos.board_key != self.board_key:
            return False
        return (0 <= self.x <= bound_pos.x and
                0 <= self.y <= bound_pos.y)

    def neighbors(self, *args: int) -> list['Position']:
        """
        按照欧几里得距离从小到大逐层扩散，筛选范围由距离平方控制（不包含当前位置）。

        调用方式（类似 range）：
            neighbors(end_layer)
                返回所有欧几里得距离 ≤ √end_layer 的位置（从第 1 层开始）。
            neighbors(start_layer, end_layer)
                返回所有欧几里得距离 ∈ [√start_layer, √end_layer] 的位置。

        :param args: 一个或两个整数
            - 若提供一个参数 end_layer，视为从 √1 到 √end_layer。
            - 若提供两个参数 start_layer 和 end_layer，视为从 √start_layer 到 √end_layer。
            - 参数非法（数量不为 1 或 2，或值非法）时返回空列表。

        :return: 位置列表，按距离从近到远排序。
        """

        # 解析参数
        if len(args) == 1:
            low, high = 1, args[0]
        elif len(args) == 2:
            low, high = args
        else:
            return []

        # 处理无效参数
        if high < low:
            return []

        x0, y0 = self.x, self.y
        directions = [(dx, dy) for dx in (-1, 0, 1)
                      for dy in (-1, 0, 1) if (dx, dy) != (0, 0)]

        heap = []  # 最小堆存储 (距离平方, x, y)
        visited = {(x0, y0)}
        result = []

        # 处理包含自身的情况 (距离平方=0)
        if low <= 0 <= high:
            result.append(self.clone())

        # 初始化邻居
        for dx, dy in directions:
            nx, ny = x0 + dx, y0 + dy
            d_sq = (nx - x0) ** 2 + (ny - y0) ** 2
            if d_sq <= high:
                heapq.heappush(heap, (d_sq, nx, ny))
                visited.add((nx, ny))

        # 遍历所有可达位置
        while heap:
            d_sq, x, y = heapq.heappop(heap)

            # 检查是否在目标范围内
            if low <= d_sq <= high:
                result.append(Position(x, y, self.board_key))

            # 扩展新位置
            for dx, dy in directions:
                nx, ny = x + dx, y + dy
                if (nx, ny) in visited:
                    continue

                visited.add((nx, ny))
                new_d_sq = (nx - x0) ** 2 + (ny - y0) ** 2

                # 仅考虑距离平方未超过上限的位置
                if new_d_sq <= high:
                    heapq.heappush(heap, (new_d_sq, nx, ny))

        return result


class Board(AbstractBoard):
    """
    通过实现
    """
    name = "Board1"
    version = 1

    def __init__(self, size: tuple = None, code: bytes = None):
        self._model = None
        self.board_data = dict()

        if code is None:
            if size is None:
                raise ValueError("board size Undefined")
            self.generate_board(MASTER_BOARD, size=size)
            self.board_data[MASTER_BOARD]["config"]["row_col"] = True
            self.board_data[MASTER_BOARD]["config"]["interactive"] = True
            self.board_data[MASTER_BOARD]["config"]["VALUE"] = VALUE_QUESS
            self.board_data[MASTER_BOARD]["config"]["MINES"] = MINES_TAG
            return
        for chunks in code.split(b"\xff\xff"):
            board_key, chunks = chunks.split(b"\xff", 1)
            board_key = board_key.decode("ascii")
            self.generate_board(board_key, code=chunks)

    def __call__(
            self, target: Union[str, None] = "always",
            mode: str = "object",
            key: str = None,
    ) -> Generator[
        Tuple[
            'Position',
            Union[
            'AbstractClueValue',
            'AbstractMinesValue',
            str, IntVar, bool, None
            ]],
        Any, None
    ]:
        """
        被调用时循环返回目标值

        @:param
            target (str): 遍历目标类型。可选值：
                - "C": 线索 (Clue)
                - "F": 雷 (Mines)
                - "N": 未定义或未翻开
                - "always": 默认，遍历所有

            mode (str): 返回的目标类型, 可选值:
                - "object":     存储在board内的实例对象
                - "type":       对象的类型('C', 'F', 'N')
                - "variable":   变量对象
                - "dye":        染色bool

        @:return
            当前位置与选择的值。
        """
        if key is None:
            for key in self.get_interactive_keys():
                for i in self(target=target, mode=mode, key=key):
                    yield i
        else:
            size = self.board_data[key]["config"]["size"]
            for posx in range(size[0]):
                for posy in range(size[1]):
                    pos = Position(posx, posy, key)
                    if pos in self.get_config(pos.board_key, "mask"):
                        continue
                    pos_type = self.get_type(pos)

                    # 检查是否符合目标类型
                    if target == "always" or pos_type in target:
                        if mode == "object":
                            yield pos, self.get_value(pos)
                        elif mode == "obj":
                            yield pos, self.get_value(pos)
                        elif mode == "type":
                            yield pos, pos_type
                        elif mode == "var":
                            yield pos, self.get_variable(pos)
                        elif mode == "variable":
                            yield pos, self.get_variable(pos)
                        elif mode == "dye":
                            yield pos, self.get_dyed(pos)
                        elif mode == "none":
                            yield pos, None

    def has(self, target: str, key: str = None) -> bool:
        if key not in self.get_board_keys() + [None]:
            return False
        for pos, type_obj in self(mode="type", key=key):
            if type_obj == target:
                return True
        return False

    def get_model(self):
        if self._model is None:
            self._model = cp_model.CpModel()
            for _key in self.board_data:
                _size = self.board_data[_key]["config"]["size"]
                if "variable" in self.board_data[_key]:
                    del self.board_data[_key]["variable"]
                self.board_data[_key]["variable"] = \
                    [[self._model.NewBoolVar(f"var({self.get_pos(x, y, _key)})")
                      for y in range(_size[1])]
                     for x in range(_size[0])]
                get_logger().trace(f"构建新变量:{self.board_data[_key]['variable']}")
        return self._model

    def generate_board(
            self, board_key: str,
            size: tuple = (),
            code: bytes = None,
            true_tag: "AbstractValue" = VALUE_CROSS,
            false_tag: "AbstractValue" = VALUE_CIRCLE,
    ) -> None:
        """
        创建一个新的题板
        :param board_key: 题板名称
        :param size: 题板的尺寸 (与code二选一)
        :param code: 题板的代码 (与size二选一)
        :param true_tag: 题板默认非雷对象
        :param false_tag:题板默认雷对象
        """
        if board_key in self.board_data:
            return
        flag_byte = 0
        mask = 0
        if code is not None:
            config, mask, ture_code, false_code, *code\
                = code.split(b"\xff", 4)
            code = b''.join(code)
            *size, flag_byte = config
            size = (size[0], size[1])
            mask = decode_bytes_7bit(mask)
            true_tag = get_value(pos=POSITION_TAG, code=ture_code)
            false_tag = get_value(pos=POSITION_TAG, code=false_code)
        data = dict()
        self.board_data[board_key] = data
        # 配置初始化
        data["config"] = {
            "size": size,
            "VALUE": true_tag,
            "MINES": false_tag,
            "mask": [],
        }
        for key in self.CONFIG_FLAGS:
            data["config"].update({key: False})

        data["obj"] = [[None for _ in range(size[0])] for _ in range(size[1])]
        data["type"] = [["N" for _ in range(size[0])] for _ in range(size[1])]
        data["dye"] = [[False for _ in range(size[0])] for _ in range(size[1])]

        if code is None:
            return
        positions = [pos for pos, _ in self(key=board_key, mode="none")]
        for pos in positions[::-1]:
            if mask & 1:
                self.set_mask(pos)
            mask >>= 1

        for i, key in enumerate(self.CONFIG_FLAGS[::-1]):
            data["config"].update({key: bool(flag_byte & (1 << i))})

        codes = code.split(b"\xff")

        code_queue = []

        for part in codes:
            if not part:
                continue
            if part[0] == 0:
                count = int(part[1])
                code_queue.extend([b'_'] * count)
            else:
                code_queue.append(part)

        for pos, _ in self(key=board_key):
            code = code_queue.pop(0)
            if code[0] == 35:
                self.set_dyed(pos, True)
                code = code[1:]
            if code == b'_':
                continue
            value = get_value(pos, code)
            if value is not None:
                self.set_value(pos, value)
                continue
            raise ValueError(f"unknown type{code}")

    def encode(self) -> bytes:
        """
        字节头: 尺寸
        无需换行符 初始化自动排序
        '_'表示None
        :return: 字节码
        """
        board_bytes = bytearray()
        for board_key in self.board_data:
            size = self.board_data[board_key]["config"]["size"]
            value = self.board_data[board_key]["config"]["VALUE"]
            mines = self.board_data[board_key]["config"]["MINES"]
            flags = 0
            mask = 0
            for name in self.CONFIG_FLAGS:
                flags = (flags << 1) | int(self.board_data[board_key]["config"].get(name, False))
            for posx in range(size[0]):
                for posy in range(size[1]):
                    pos = self.get_pos(posx, posy, board_key)
                    mask <<= 1
                    if pos is None:
                        mask |= 1
            board_bytes.extend(board_key.encode("ascii") + b"\xff")
            board_bytes.extend(bytes([size[0], size[1], flags, 255]))
            board_bytes.extend(encode_int_7bit(mask) + bytes([255]))
            board_bytes.extend(value.type() + b"|" + value.code())
            board_bytes.extend(bytes([255]))
            board_bytes.extend(mines.type() + b"|" + mines.code())
            # key | sizex | sizey | config
            for pos, obj in self(key=board_key):
                board_bytes.extend(b"\xff")
                if self.get_dyed(pos):
                    board_bytes.extend(b"#")
                if obj is None:
                    board_bytes.extend(b"_")
                else:
                    code = obj.code()
                    if b"\xff" in code:
                        get_logger().error(f"{obj.type().decode()}中编码出现\\xff")
                        raise ValueError(r"code contains forbidden byte: \xff")
                    board_bytes.extend(obj.type() + b"|" + code)
            board_bytes.extend(b"\xff\xff")
        # 只用split(b"\xff_")切分
        parts = board_bytes.split(b"\xff_")

        # 处理连续 \xff_ 的次数
        encoded_bytes = bytearray()
        i = 0
        while i < len(parts):
            if i > 0:
                # 统计连续 \xff_ 的次数
                count = 1
                # 看后续parts里是否以空字节开头来判断是否连续（split后的空串）
                # 但这里由于只分割 \xff_，连续情况只能靠检查下一个part是否空
                while i + count < len(parts) and parts[i + count - 1] == b'':
                    count += 1
                # 输出 \xff + 数字（表示连续多少个 \xff_）
                while count > 254:
                    encoded_bytes.extend(b"\xff\x00" + bytes([254]))
                    count -= 254
                if count > 0:
                    encoded_bytes.extend(b"\xff\x00" + bytes([count]))
                i += count - 1
            encoded_bytes.extend(parts[i])
            i += 1
        return encoded_bytes[:-2]

    def boundary(self, key=MASTER_BOARD) -> "Position":
        if key not in self.get_board_keys():
            return Position(-1, -1, key)
        size = self.board_data[key]["config"]["size"]
        return Position(size[0] - 1, size[1] - 1, key)

    def is_valid(self, pos: 'AbstractPosition') -> bool:
        if pos in self.get_config(pos.board_key, "mask"):
            return False
        return super().is_valid(pos)

    @staticmethod
    def type_value(value) -> str:
        # 查看value的类型
        if value is None:
            return "N"
        elif isinstance(value, AbstractMinesValue):
            return "F"
        elif isinstance(value, AbstractClueValue):
            return "C"
        get_logger().error(f"unknown type: value{value}, type{type(value)}")
        raise ValueError(f"unknown type: {value}, type{type(value)}")

    def get_type(self, pos: 'Position') -> str:
        key = pos.board_key
        if self.is_valid(pos):
            return self.board_data[key]["type"][pos.y][pos.x]
        return ""

    def get_value(self, pos: 'Position') -> Union['AbstractClueValue', 'AbstractMinesValue', None]:
        key = pos.board_key
        if self.is_valid(pos):
            return self.board_data[key]["obj"][pos.y][pos.x]
        return None

    def set_value(self, pos: 'Position', value):
        key = pos.board_key
        if self.is_valid(pos):
            self.board_data[key]["obj"][pos.y][pos.x] = value
            self.board_data[key]["type"][pos.y][pos.x] = self.type_value(value)

    def get_dyed(self, pos: 'Position') -> bool | None:
        key = pos.board_key
        if self.is_valid(pos):
            return self.board_data[key]["dye"][pos.y][pos.x]

    def set_dyed(self, pos: 'Position', dyed: bool):
        key = pos.board_key
        if self.is_valid(pos):
            self.board_data[key]["dye"][pos.y][pos.x] = dyed

    def get_variable(self, pos: 'Position') -> IntVar | None:
        key = pos.board_key
        self.get_model()
        if self.is_valid(pos):
            return self.board_data[key]["variable"][pos.x][pos.y]

    def clear_variable(self):
        for key in self.board_data.keys():
            if "variable" in self.board_data[key]:
                del self.board_data[key]["variable"]
        self._model = None
        gc.collect()

    def get_config(self, board_key, config_name):
        if board_key not in self.board_data:
            return None
        return self.board_data[board_key]["config"][config_name]

    def set_config(self, board_key, config_name, value: bool):
        if board_key not in self.board_data:
            return None
        self.board_data[board_key]["config"][config_name] = value

    def set_mask(self, pos):
        if not self.is_valid(pos):
            return
        self.get_config(pos.board_key, "mask").append(pos)

    def get_row_pos(self, pos: 'Position') -> List["Position"]:
        bound = self.boundary(pos.board_key)
        _pos = pos.clone()
        pos_list = [_pos]
        while True:
            _pos = _pos.left()
            if not _pos.in_bounds(bound):
                break
            pos_list.append(_pos)
        _pos = pos.clone()
        pos_list = pos_list[::-1]
        while True:
            _pos = _pos.right()
            if not _pos.in_bounds(bound):
                break
            pos_list.append(_pos)
        return pos_list

    def get_col_pos(self, pos: 'Position') -> List["Position"]:
        bound = self.boundary(pos.board_key)
        _pos = pos.clone()
        pos_list = [_pos]
        while True:
            _pos = _pos.up()
            if not _pos.in_bounds(bound):
                break
            pos_list.append(_pos)
        _pos = pos.clone()
        pos_list = pos_list[::-1]
        while True:
            _pos = _pos.down()
            if not _pos.in_bounds(bound):
                break
            pos_list.append(_pos)
        return pos_list

    def get_pos(self, x: int, y: int, key=MASTER_BOARD) -> Union['Position', None]:
        size = self.board_data[key]["config"]["size"]
        if -size[0] < x < size[0] and -size[1] < y < size[1]:
            x = x if x >= 0 else size[0] + x
            y = y if y >= 0 else size[1] + y
            pos = Position(x, y, key)
            if self.is_valid(pos):
                return pos
        return None

    def get_pos_box(self, pos1: "AbstractPosition", pos2: "AbstractPosition") -> List["AbstractPosition"]:
        if pos1.board_key != pos2.board_key:
            return []
        if not (self.in_bounds(pos1) and self.in_bounds(pos2)):
            return []
        x_min, x_max = sorted([pos1.x, pos2.x])
        y_min, y_max = sorted([pos1.y, pos2.y])

        result = []
        for y in range(y_min, y_max + 1):
            for x in range(x_min, x_max + 1):
                result.append(self.get_pos(x, y, key=pos1.board_key))
        return result

    def batch(self, positions: List['Position'],
              mode: str, drop_none: bool = False) -> List[Any]:
        result = []
        for pos in positions:
            if drop_none and not self.in_bounds(pos):
                continue
            if mode == "object":
                result.append(self.get_value(pos))
            elif mode == "obj":
                result.append(self.get_value(pos))
            elif mode == "variable":
                result.append(self.get_variable(pos))
            elif mode == "var":
                result.append(self.get_variable(pos))
            elif mode == "type":
                result.append(self.get_type(pos))
            elif mode == "dye":
                result.append(self.get_dyed(pos))
            else:
                raise ValueError(f"Unsupported mode: {mode}")
        return result

    def clear_board(self):
        for key in self.board_data:
            data = self.board_data[key]
            size = data["config"]["size"]
            data["obj"] = [[None for _ in range(size[0])] for _ in range(size[1])]
            data["type"] = [["N" for _ in range(size[0])] for _ in range(size[1])]
            self.clear_variable()

    def get_board_keys(self) -> list[str]:
        return list(self.board_data.keys())

    def show_board(self, show_tag: bool = False):
        r = ""
        for key in self.board_data:
            size = self.board_data[key]["config"]["size"]
            if len(self.board_data.keys()) > 1:
                r += key + "\n"
            for i in range(size[0]):
                for j in range(size[1]):
                    pos = self.get_pos(i, j, key)
                    if pos is None:
                        r += "\t\t" if show_tag else "\t"
                        continue
                    value = self[pos]
                    if value is None:
                        r += "______" if show_tag else "___"
                    else:
                        r += str(value) + ("_" + value.type().decode() if show_tag else "")
                    r += "\t"
                r += "\n"
            r += "\n\n"
        return r[:-2]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/06 05:43
# @Author  : Wu_RH
# @FileName: generate_test.py
import os
import threading
import time

from minesweepervariants.impl.summon import Summon
from minesweepervariants.utils.image_create import draw_board
from minesweepervariants.utils.impl_obj import get_seed
from minesweepervariants.utils.tool import get_logger, get_random

from minesweepervariants.config.config import DEFAULT_CONFIG, PUZZLE_CONFIG

# ==== 获取默认值 ====
CONFIG = {}
CONFIG.update(DEFAULT_CONFIG)
CONFIG.update(PUZZLE_CONFIG)


def main(
        log_lv: str,  # 日志等级
        seed: int,  # 随机种子
        size: tuple[int, int],  # 题板尺寸
        total: int,  # 总雷数
        rules: list[str],  # 规则id集合
        dye: str,  # 染色规则
        mask_dye: str,   # 异形题板
        board_class: str,  # 题板的名称
        unseed: bool
):
    logger = get_logger(log_lv=log_lv)
    get_random(seed, new=True)
    attempt_index = 0
    s = Summon(size=size, total=total, rules=rules, board=board_class, mask=mask_dye, dye=dye)
    if unseed:
        s.unseed = unseed
    total = s.total
    logger.info(f"total mines: {total}")
    _board = None
    while True:
        attempt_index += 1
        logger.info(f"尝试第{attempt_index}次minesweepervariants..", end="\r")
        get_random(seed, new=True)
        a_time = time.time()
        _board = s.summon_board()
        if _board is None:
            continue
        logger.info(f"<{attempt_index}>生成用时:{(time_used := time.time() - a_time)}s")
        logger.info(f"总雷数: {total}")
        logger.info("\n" + _board.show_board())

        rule_text = ""
        for rule in rules:
            rule_text += "[" + (rule.split(CONFIG['delimiter'])[0] if
                                CONFIG['delimiter'] in rule else rule) + "]"
        if rule_text == "":
            rule_text = "[V]"
        if dye:
            rule_text += f"[@{dye}]"
        rule_text += f"{size[0]}x{size[1]}"

        if not os.path.exists(CONFIG["output_path"]):
            os.mkdir(CONFIG["output_path"])

        with open(os.path.join(CONFIG["output_path"], "demo.txt"), "a", encoding="utf-8") as f:
            f.write("\n" + ("=" * 100) + "\n\n生成时间" + logger.get_time() + "\n")
            f.write(f"生成用时:{time_used}s\n")
            f.write(f"总雷数: {total}\n")
            f.write(f"种子: {get_seed()}\n")
            f.write(rule_text)
            f.write("\n"+_board.show_board())

            f.write(f"\n题板: img -c {_board.encode().hex()} ")
            f.write(f"-r \"{rule_text}-R{total}")
            if unseed:
                f.write(" ")
            else:
                f.write(f"-{get_seed()}\" ")
            f.write("-o answer\n")

        def d():
            draw_board(board=_board, cell_size=100, output="answer",
                       bottom_text=rule_text + f"-R{total}-{get_seed()}\n")
        threading.Thread(target=d, daemon=True).start()

        logger.info("\n\n" + "=" * 20 + "\n")
        logger.info("\n生成时间" + logger.get_time() + "\n")
        logger.info(f"生成用时:{time_used}s\n")
        logger.info(f"总雷数: {total}\n")
        logger.info("\n" + _board.show_board() + "\n")

        input("检查完毕后输入回车继续尝试 使用ctrl+c终止进程\r")
        attempt_index = 0



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/12 23:04
# @Author  : Wu_RH
# @FileName: 3D1X.py
"""
[3D1X'] 三维十字'(Cross)：线索表示半径为 1 的十字范围内的雷数(上下左右前后)
"""
from typing import List

from .....abs.board import AbstractPosition, AbstractBoard
from .. import Abstract3DClueRule
from .....abs.Rrule import AbstractClueValue


class Rule3D1X(Abstract3DClueRule):
    name = ["3D1X'", "三维十字'"]
    doc = "线索表示半径为 1 的十字范围内的雷数(上下左右前后)"

    def __init__(self, board: AbstractBoard, data: str = None):
        super().__init__(board, data)
        # print(board.show_board())

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            nei = Rule3D1X.pos_neighbors(board, pos, 1)
            value = board.batch(nei, mode="type").count("F")
            obj = Value3D1X(pos, bytes([value]))
            board.set_value(pos, obj)

        return board


class Value3D1X(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.pos = pos

    def __repr__(self) -> str:
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        return Rule3D1X.pos_neighbors(board, self.pos, 1)

    @classmethod
    def type(cls) -> bytes:
        return Rule3D1X.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        nei = Rule3D1X.pos_neighbors(board, self.pos, 1)

        nei_vars = board.batch(nei, mode="variable", drop_none=True)

        model.Add(sum(nei_vars) == self.value).OnlyEnforceIf(s)



"""
[2Q'] 章鱼 (Octopus)：线索表示八方向上最近的2格有雷的方向数量。
"""

from typing import List

from minesweepervariants.utils.impl_obj import MINES_TAG, VALUE_QUESS
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.tool import get_logger

def pos_shift(
        board: AbstractBoard,
        pos: AbstractPosition,
        index: int
):
    shift = [
        (1, 1), (1, 0), (1, -1),
        (0, 1), (0, -1),
        (-1, 1), (-1, 0), (-1, -1)
    ][index]
    positions = []
    for _ in range(2):
        if board.is_valid(_pos := pos.shift(shift[0], shift[1])):
            positions.append(_pos)
            shift = (
                0 if shift[0] == 0 else (shift[0] + 1 if shift[0] > 0 else shift[0] - 1),
                0 if shift[1] == 0 else (shift[1] + 1 if shift[1] > 0 else shift[1] - 1),
            )
    return positions

class Rule2QPrime(AbstractClueRule):
    name = ["2Q'", "章鱼", "Octopus"]
    doc = "线索表示八方向上最近的2格有雷的方向数量"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            value = 0
            for i in range(8):
                if "F" in board.batch(pos_shift(board, pos, i), mode="type"):
                    value += 1
            board.set_value(pos, Value2QPrime(pos, count=value))
        return board

class Value2QPrime(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            self.count = code[0]
        else:
            self.count = count
        self.neighbor = None

    def __repr__(self):
        return str(self.count)

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        positions = []

        for i in range(8):
            pos_list = pos_shift(board, self.pos, i)
            type_list = board.batch(pos_list, mode="type")
            if "F" in type_list:
                positions.append(pos_list[type_list.index("F")])
            else:
                positions.extend(pos_list)

        return positions

    @classmethod
    def type(cls) -> bytes:
        return Rule2QPrime.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])
    
    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self.pos)

        tmp_vars = []
        for index in range(8):
            tmp = model.NewBoolVar("tmp")
            var_list = board.batch(pos_shift(board, self.pos, index), mode="variable")
            model.AddBoolOr(var_list).OnlyEnforceIf(tmp)
            model.Add(sum(var_list) == 0).OnlyEnforceIf(tmp.Not())
            tmp_vars.append(tmp)
        model.Add(sum(tmp_vars) == self.count).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/04 07:39
# @Author  : Wu_RH
# @FileName: 4V.py
"""
[4V2E'2I']映射自指残缺(2X'plus+2E'+2I'): 字母X是两个题板中相同位置为中心其中一个的3x3区域中某?格的雷总数为N。则该对应位置所属的题板在标有X=N的位置必然是雷, 且?格的位置全局共享
(注:生成不知道是不是概率问题 会出现大量的生成失败 不过也不是人玩的反正 加个-r估计会好)
"""
from .....abs.Rrule import AbstractClueRule, AbstractClueValue
from .....abs.board import AbstractBoard, AbstractPosition, MASTER_BOARD
from .....utils.impl_obj import VALUE_QUESS, MINES_TAG, VALUE_CIRCLE, VALUE_CROSS
from .....utils.tool import get_random, get_logger
from . import BOARD_NAME_4V

NAME = BOARD_NAME_4V
NAME_4V_2Ip = "4V2I'"
ALPHABET = "ABCDEFGHIJ"


class Rule4V2Ep2Ip(AbstractClueRule):
    name = ["4V2E'2I'", "自指残缺映射"]
    doc = "字母X是两个题板中相同位置为中心其中一个的3x3区域中某?格的雷总数为N。则该对应位置所属的题板在标有X=N的位置必然是雷, 且?格的位置全局共享"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        size = (board.boundary().x + 1, board.boundary().y + 1)
        board.generate_board(NAME, size)
        board.set_config(NAME, "interactive", True)
        board.set_config(NAME, "row_col", True)
        board.set_config(NAME, "VALUE", VALUE_QUESS)
        board.set_config(NAME, "MINES", MINES_TAG)
        board.set_config(NAME, "pos_label", True)
        board.set_config(MASTER_BOARD, "pos_label", True)
        board.generate_board(NAME_4V_2Ip, (3, 3))

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        def apply_offsets(_pos: AbstractPosition):
            nonlocal offsets
            result = []
            for dpos in offsets:
                result.append(_pos.deviation(dpos))
            return result

        random = get_random()

        pos_list = [pos for pos, _ in board("N", key=NAME_4V_2Ip)]
        pos_list = random.sample(pos_list, int(random.random() * 6 + 3))
        offsets = []
        for pos in pos_list:
            board[pos] = VALUE_CIRCLE
            offsets.append(pos.up().left())
        for pos, _ in board("N", key=NAME_4V_2Ip):
            board[pos] = VALUE_CROSS

        letter_map_a = {i: [] for i in range(10)}
        for pos, _ in board("F", key=MASTER_BOARD):
            if pos.y > 9:
                continue
            letter = ALPHABET[pos.y]
            if pos.x not in letter_map_a:
                letter_map_a[pos.x] = []
            letter_map_a[pos.x].append(letter)

        letter_map_b = {i: [] for i in range(10)}
        for pos, _ in board("F", key=NAME):
            if pos.y > 9:
                continue
            letter = ALPHABET[pos.y]
            if pos.x not in letter_map_b:
                letter_map_b[pos.x] = []
            letter_map_b[pos.x].append(letter)

        for pos, _ in board(key=MASTER_BOARD):
            neighbors_list = []
            for _key in [MASTER_BOARD, NAME]:
                _pos = pos.clone()
                _pos.board_key = _key
                neighbors_list.append(apply_offsets(_pos))
            values = [board.batch(positions, mode="type").count("F") for positions in neighbors_list]
            values[0] = ALPHABET.index(random.choice(letter_map_a[values[0]])) if letter_map_a[values[0]] else -1
            values[1] = ALPHABET.index(random.choice(letter_map_b[values[1]])) if letter_map_b[values[1]] else -1
            r_value = 0 if random.random() > 0.7 else 1
            _pos.board_key = MASTER_BOARD
            if board.get_type(_pos) != "F":
                obj = Value4V2Ep2Ip(pos=_pos, code=bytes([values[r_value]])) if values[r_value] != -1 else VALUE_QUESS
                board.set_value(_pos, obj)
            _pos.board_key = NAME
            if board.get_type(_pos) != "F":
                obj = Value4V2Ep2Ip(pos=_pos, code=bytes([values[1 - r_value]])) \
                    if values[1 - r_value] != -1 else VALUE_QUESS
                board.set_value(_pos, obj)
        return board

    def suggest_total(self, info: dict):
        ub = 0
        for key in info["interactive"]:
            size = info["size"][key]
            ub += size[0] * size[1]
        info["soft_fn"](ub * 0.4)

    def init_clear(self, board: 'AbstractBoard'):
        for pos, obj in board(mode="object", key=NAME_4V_2Ip):
            board[pos] = None


class Value4V2Ep2Ip(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.pos = pos

    @classmethod
    def type(cls) -> bytes:
        return Rule4V2Ep2Ip.name[0].encode("ascii")

    def __repr__(self) -> str:
        return f"{ALPHABET[self.value]}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        positions = []
        for pos_key in [MASTER_BOARD, BOARD_NAME_4V]:
            self_pos = self.pos.clone()
            self_pos.board_key = pos_key
            for pos, _ in board("NF", key=NAME_4V_2Ip):
                _pos = self_pos.deviation(pos.shift(1, -1))
                if board.in_bounds(_pos):
                    positions.append(_pos)
        return positions

    def create_constraints(self, board: 'AbstractBoard', switch):
        # 初始化模型
        model = board.get_model()
        s = switch.get(model, self)
        # 初始化日志
        logger = get_logger()

        sum_var = []
        for key in [MASTER_BOARD, NAME]:
            _pos = self.pos.clone()
            _pos.board_key = key

            b = model.NewBoolVar(f"[{Rule4V2Ep2Ip.name}]tmp")
            sum_var.append(b)

            # 初始化位置对象 位于X列
            pos = board.get_pos(0, self.value, key)
            # 获取该列的所有位置
            line = board.get_col_pos(pos)
            # 获取该列的所有变量
            line_vars = board.batch(line, mode="variable")

            # 初始化对照表
            neighbors = []
            for pos2, obj in board(key=NAME_4V_2Ip):
                # 题板上的位置和共享的偏移位置
                _positions = [_pos.deviation(pos2).up().left(), pos2]
                # 第一个为题板对应的变量 第二个为偏移的变量
                if not board.in_bounds(_positions[0]):
                    continue
                neighbors.append(board.batch(_positions, mode="variable"))

            # 初始化和值
            sum_vers = []
            for var_to_sum, cond in neighbors:
                # 初始化临时变量
                tmp = model.NewBoolVar(f"included_if_{_pos}_{var_to_sum}")
                # 如果偏移变量为真 那么tmp为题板的值
                model.Add(tmp == var_to_sum).OnlyEnforceIf([cond, b, s])
                # 如果偏移变量为假 那么tmp为0
                model.Add(tmp == 0).OnlyEnforceIf([cond.Not(), b, s])
                sum_vers.append(tmp)
                logger.trace(f"[4V2E'2I'] new tempVar: {tmp} = if {cond} -> {var_to_sum}")

            for index in range(min(8, len(line_vars))):
                # 获取该列的X=index的变量
                var = line_vars[index]
                # 如果变量为真 那么sum应该相对 反之亦然
                model.Add(sum(sum_vers) != index).OnlyEnforceIf([var.Not(), b, s])
                logger.trace(f"[4V2E'2I'] sum_tmp != {index} only if not {var}")

        model.AddBoolOr(sum_var).OnlyEnforceIf(s)
        # model.Add(sum(sum_var) == 0)

    def code(self) -> bytes:
        return bytes([self.value])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/23 22:35
# @Author  : Wu_RH
# @FileName: root_shape.py
from abc import ABC
from typing import Tuple, List, Optional

from minesweepervariants.abs.Rrule import AbstractClueRule
from minesweepervariants.abs.Mrule import AbstractMinesClueRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.abs.rule import AbstractRule
from minesweepervariants.impl.impl_obj import get_rule
from minesweepervariants.impl.rule.Rrule.sharp import RuleSharp as ClueSharp
from minesweepervariants.impl.rule.Mrule.sharp import RuleSharp as MinesSharp
from minesweepervariants.impl.summon.solver import Switch
from minesweepervariants.utils.tool import get_logger


class AbstractMinesSharp(AbstractMinesClueRule, ABC):
    def __init__(self, rules: list[str], board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        rule_list = []
        for name in rules:
            try:
                rule_list.append(get_rule(name)(board=board, data=None))
            except Exception as e:
                get_logger().warn(f"加载规则:[{name}]失败")
                get_logger().warn(f"ERROR: {e}")
        self.shape_rule = MinesSharp(board=board, data=rule_list)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        return self.shape_rule.fill(board=board)

    def init_board(self, board: 'AbstractBoard'):
        return self.shape_rule.init_board(board=board)

    def init_clear(self, board: 'AbstractBoard'):
        return self.shape_rule.init_clear(board=board)

    def combine(self, rules: List[Tuple['AbstractRule', Optional[str]]]):
        return self.shape_rule.combine(rules=rules)

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        return self.shape_rule.create_constraints(board=board, switch=switch)

    def suggest_total(self, info: dict):
        return self.shape_rule.suggest_total(info=info)


class AbstractClueSharp(AbstractClueRule, ABC):
    def __init__(self, rules: list[str], board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        rule_list = []
        for name in rules:
            try:
                rule_list.append(get_rule(name)(board=board, data=None))
            except Exception as e:
                get_logger().warn(f"加载规则:[{name}]失败")
                get_logger().warn(f"ERROR: {e}")
        self.shape_rule = ClueSharp(board=board, data=rule_list)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        return self.shape_rule.fill(board=board)

    def init_board(self, board: 'AbstractBoard'):
        return self.shape_rule.init_board(board=board)

    def init_clear(self, board: 'AbstractBoard'):
        return self.shape_rule.init_clear(board=board)

    def combine(self, rules: List[Tuple['AbstractRule', Optional[str]]]):
        return self.shape_rule.combine(rules=rules)

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        return self.shape_rule.create_constraints(board=board, switch=switch)

    def suggest_total(self, info: dict):
        return self.shape_rule.suggest_total(info=info)




#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/06/03 18:38
# @Author  : Wu_RH
# @FileName: Rrule.py

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING, List, Dict

from .rule import AbstractRule, AbstractValue
from ..utils.image_create import get_text, get_image, get_dummy, get_col
from ..utils.web_template import Number

if TYPE_CHECKING:
    from minesweepervariants.abs.board import AbstractBoard, AbstractPosition


class AbstractClueRule(AbstractRule):
    """
    数字线索规则
    """

    @abstractmethod
    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        """
        填充所有None为规则线索对象
        :param board: 题板
        :return: 题板
        """
        ...


class AbstractClueValue(AbstractValue, ABC):
    """
    线索格数字对象类
    """

    def __repr__(self) -> str:
        """
        当前值在展示时候的显示字符串
        :return: 显示的字符串
        """
        return "?"

    def compose(self, board) -> Dict:
        """
        返回一个可渲染对象列表
        默认使用__repr__
        """
        return get_col(
            get_dummy(height=0.3),
            get_text(self.__repr__()),
            get_dummy(height=0.3),
        )

    def web_component(self, board) -> Dict:
        """
        返回一个可渲染对象列表
        默认使用__repr__
        """
        if "compose" in type(self).__dict__:
            return self.compose(board)
        return Number(self.__repr__())


# --------实例类-------- #


class ValueQuess(AbstractClueValue):
    """
    问号类(线索非雷)
    """

    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)

    def __repr__(self):
        return "?"

    @classmethod
    def type(cls) -> bytes:
        return b"?"

    def code(self) -> bytes:
        return b""

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        return []


class ValueCross(AbstractClueValue):
    """
    副板的叉号
    """

    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)

    def __repr__(self):
        return "X"

    def web_component(self, board) -> Dict:
        return get_image("cross")

    def compose(self, board) -> Dict:
        return get_image("cross")

    @classmethod
    def type(cls) -> bytes:
        return b"X"

    def code(self) -> bytes:
        return b""



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:13
# @Author  : Wu_RH
# @FileName: 1C.py
"""
[1K1C] 马步八连通 (Knight-Connected)：雷区域马步连通
"""
from ....abs.Lrule import AbstractMinesRule
from ....utils.impl_obj import get_total

from .connect import connect


class Rule1C(AbstractMinesRule):
    name = ["1K1C", "马步八连通", "Knight-Connected"]
    doc = "雷区域马步连通"

    def create_constraints(self, board, switch):
        model = board.get_model()
        connect(
            ub=get_total() // 2 + 1,
            model=model,
            board=board,
            connect_value=1,
            nei_value=(5, 5),
            switch=switch.get(model, self),
        )



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/10 03:17
# @Author  : Wu_RH
# @FileName: 2G'.py
"""
[2G'] 三连块 (Group')：所有四连通雷区域的面积为 3
"""
# (在提示的表现似乎有问题)
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule2Gp(AbstractMinesRule):
    name = ["2G'", "三联块", "Group'"]
    doc = "所有四连通雷区域的面积为3"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="variable"):
            nei = board.batch(pos.neighbors(1), mode="variable", drop_none=True)
            model.Add(sum(nei) < 3).OnlyEnforceIf([var, s])
            model.Add(sum(nei) > 0).OnlyEnforceIf([var, s])
            tmp = model.NewBoolVar("tmp")
            model.Add(sum(nei) == 2).OnlyEnforceIf([tmp, s])
            model.Add(sum(nei) != 2).OnlyEnforceIf([tmp.Not(), s])
            for _pos in pos.neighbors(1):
                if not board.is_valid(_pos):
                    continue
                _var = board.get_variable(_pos)
                nei = board.batch(_pos.neighbors(1), mode="variable", drop_none=True)
                model.Add(sum(nei) == 1).OnlyEnforceIf([_var, var, tmp, s])
                model.Add(sum(nei) == 2).OnlyEnforceIf([_var, var, tmp.Not(), s])

    def suggest_total(self, info: dict):
        def hard_constraint(m, total):
            m.AddModuloEquality(0, total, 3)

        info["hard_fns"].append(hard_constraint)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2025/06/07 19:40
# @Author  : Wu_RH
# @FileName: solver.py
import os
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from ortools.sat.python.cp_model import IntVar

from ...abs.rule import AbstractRule, AbstractValue
from ortools.sat.python import cp_model
from collections import defaultdict
from typing import Dict, List, Tuple, Union, Optional

from ...utils.tool import get_logger

from ...abs.Mrule import AbstractMinesClueRule
from ...abs.Rrule import AbstractClueRule
from ...abs.board import AbstractBoard, AbstractPosition
from ...abs.Lrule import MinesRules, Rule0R
from ..impl_obj import ModelGenerateError

from minesweepervariants.config.config import DEFAULT_CONFIG

# ==== 获取默认值 ====
CONFIG = {}
CONFIG.update(DEFAULT_CONFIG)


def get_solver(b: bool):
    solver = cp_model.CpSolver()
    solver.parameters.random_seed = 42  # 启发式多样性
    solver.parameters.num_search_workers = CONFIG["workes_number"]  # 多线程并行搜索
    solver.parameters.search_branching = cp_model.AUTOMATIC_SEARCH
    solver.parameters.linearization_level = 2  # 启用线性化加速
    # solver.parameters.cp_model_presolve = b
    solver.parameters.use_optional_variables = True
    solver.parameters.randomize_search = True
    if CONFIG["timeout"] > 0 and b:
        solver.parameters.max_time_in_seconds = CONFIG["timeout"]  # 时间限制
    return solver


class Switch:

    def __init__(self):
        # 对象名称到索引计数器的映射（用于自动生成索引）
        self.name_counter: Dict[str, int] = defaultdict(lambda: 1)

        # 存储所有创建的变量
        self.all_vars: List[cp_model.IntVar] = []

        # 对象到其所有开关的映射 {obj_str: [(index_str, var), ...]}
        self.obj_switches: Dict[str, List[Tuple[str, cp_model.IntVar]]] = defaultdict(list)

        # 注册表：存储已创建的变量 {(obj_str, index_str): var}
        self.var_registry: Dict[Tuple[str, str], cp_model.IntVar] = {}

        # 反向映射：变量索引到(obj_str, index_str)的映射
        self.var_to_key: Dict[int, Tuple[str, str]] = {}

    def __call__(self, *args, **kwargs):
        return self.get(*args, **kwargs)

    @staticmethod
    def to_str(obj: Union[AbstractRule, AbstractValue, AbstractPosition, str]) -> str:
        if isinstance(obj, AbstractRule):
            name = f"RULE|{obj.get_name()}"
        elif (
                isinstance(obj, AbstractPosition) or
                isinstance(obj, AbstractValue)
        ):
            if isinstance(obj, AbstractValue):
                pos = obj.pos
            else:
                pos = obj
            if pos is None:
                raise ValueError("pos cannot be None")
            pos = f"|{pos.x}|{pos.y}|{pos.board_key}"
            name = "POS" + pos
        elif isinstance(obj, str):
            name = obj
        else:
            raise ValueError("obj must be an Rule or Value")
        return name

    def get(
            self,
            model: cp_model.CpModel,
            obj: Union[AbstractRule, AbstractValue, AbstractPosition, str],
            index_str: Optional[str] = None
    ) -> cp_model.IntVar:
        """
        创建或获取一个开关变量

        参数:
        model: CpModel实例
        obj: 关联的对象（规则、位置或值）
        index_str: 可选，指定索引字符串。如果未提供则自动生成

        返回:
        布尔变量（如果是新创建或已存在的变量）
        """
        obj_str = self.to_str(obj)

        # 自动生成索引字符串（如果需要）
        if index_str is None:
            # 获取下一个索引并递增
            current_index = self.name_counter[obj_str]
            index_str = str(current_index)
            self.name_counter[obj_str] += 1

        # 检查是否已存在相同的变量
        key = (obj_str, index_str)
        if key in self.var_registry:
            return self.var_registry[key]

        # 创建新变量
        var = model.NewBoolVar(f"{obj_str}_{index_str}")

        # 存储变量
        self.var_registry[key] = var
        self.all_vars.append(var)
        self.obj_switches[obj_str].append((index_str, var))
        self.var_to_key[var.Index()] = (obj_str, index_str)

        # 尝试更新计数器（仅当索引是数字时）
        try:
            idx_val = int(index_str)
            self.name_counter[obj_str] = max(self.name_counter[obj_str], idx_val + 1)
        except ValueError:
            # 非数字索引时不更新计数器
            pass

        return var

    def get_switches_by_obj(
            self,
            obj: Union[AbstractRule, AbstractValue, AbstractPosition, str]
    ) -> List[Tuple[str, cp_model.IntVar]]:
        """
        获取指定对象的所有开关

        参数:
        obj: 查询的对象

        返回:
        该对象的所有开关列表，格式为[(index_str, var), ...]
        """
        obj_str = self.to_str(obj)
        return self.obj_switches.get(obj_str, [])

    def get_all_switches(
            self
    ) -> Dict[str, List[Tuple[str, cp_model.IntVar]]]:
        """
        获取所有对象及其开关

        返回:
        字典格式: {obj_str: [(index_str, var), ...]}
        """
        return dict(self.obj_switches)

    def get_all_vars(self) -> List[cp_model.IntVar]:
        """
        获取所有创建的开关变量

        返回:
        变量列表
        """
        return self.all_vars

    def get_obj_and_index_by_var(
            self,
            var: cp_model.IntVar
    ) -> Tuple[str, Optional[str]]:
        """
        根据开关变量获取对应的对象字符串和索引字符串

        参数:
        var: 开关变量

        返回:
        元组 (obj_str, index_str)

        如果找不到变量，抛出KeyError
        """
        var_index = var.Index()
        if var_index not in self.var_to_key:
            raise KeyError(f"Variable with index {var_index} not found in switch registry")
        key = self.var_to_key[var_index]
        if len(self.get_switches_by_obj(key[0])) == 1:
            return key[0], None
        return key


def solver_by_csp(
        mines_rules: MinesRules,
        clue_rule: Union[AbstractClueRule, None],
        mines_clue_rule: Union[AbstractMinesClueRule, None],
        board: AbstractBoard,
        drop_r=False,
        bool_mode=False,
        answer_board=None,
        model=None
) -> int:
    """
    返回int 0表示无解 1表示唯一解 2表示多解
    """
    # -*- csp推导 -*- #
    logger = get_logger()

    if model is None:
        logger.trace("求解器输入:\n" + board.show_board())
        logger.trace("构建新模型")
        board = board.clone()
        board.clear_variable()
        model = board.get_model()
        switch = Switch()

        # 2.获取所有规则约束
        for rule in (
                mines_rules.rules +
                [clue_rule, mines_clue_rule]
        ):
            if rule is None:
                continue
            if drop_r and isinstance(rule, Rule0R):
                continue
            rule: AbstractRule
            rule.create_constraints(board, switch)

        for key in board.get_board_keys():
            for pos, obj in board(key=key):
                if obj is None:
                    continue
                obj: AbstractValue
                obj.create_constraints(board, switch)

        # 3.获取所有变量并赋值已解完的部分
        for key in board.get_board_keys():
            for _, var in board("C", mode="variable", key=key):
                model.Add(var == 0)
                logger.trace(f"var: {var} == 0")
            for _, var in board("F", mode="variable", key=key):
                model.Add(var == 1)
                logger.trace(f"var: {var} == 1")

        model: cp_model.CpModel
        # model.AddAssumptions(switch.var_map.values())
        for switch_var in switch.get_all_vars():
            model.Add(switch_var == 1)

    # 4.获取求解器并推导
    solver = get_solver(True)

    if answer_board is not None:
        current_solution = []
        logger.trace("设置预设不同值")
        for key in board.get_interactive_keys():
            for pos, _ in board("N", key=key):
                value = 1 if answer_board.get_type(pos) == "F" else 0
                val = board.get_variable(pos)
                tmp = model.NewBoolVar(f"answer_tmp{pos}")
                current_solution.append(tmp)
                logger.trace(f"[{pos}]{val} != {value} ({answer_board.get_type(pos)})")
                model.Add(val != value).OnlyEnforceIf(tmp)
        model.AddBoolOr(current_solution)  # 新增排除当前解约束（至少有一个变量取反）

        status2 = solver.Solve(model)
        if status2 == cp_model.FEASIBLE or status2 == cp_model.OPTIMAL:
            logger.trace(f"求解器多解")
            return 2
        return 1

    status = solver.Solve(model)

    # 5.检查是否无解或者多解
    if status == cp_model.INFEASIBLE:
        logger.trace("求解器无解")
        return 0

    if status == cp_model.UNKNOWN:
        logger.trace("求解器遇到未知问题")
        return 0

    if status == cp_model.MODEL_INVALID:
        # 这个bug太玄乎了 修不好直接爆了
        # 关键他重新跑一遍就正常了 我没法debug
        logger.trace("\n" + board.show_board())
        logger.trace("求解器模型构建出现错误")

        # 输出状态码名称
        logger.trace(f"Solver status: {solver.StatusName(status)}")

        # 输出求解统计信息
        logger.trace(f"Num conflicts: {solver.NumConflicts()}")
        logger.trace(f"Num branches: {solver.NumBranches()}")
        logger.trace(f"Wall time (s): {solver.WallTime()}")

        for key in board.get_board_keys():
            logger.trace(f"board key: {key}")
            logger.trace("varlist:{}".format([(var.index, var) for _, var in board(mode="variable", key=key)]))
            logger.trace("obj:{}".format([var for _, var in board(mode="object", key=key)]))
            logger.trace("type:{}".format([var for _, var in board(mode="type", key=key)]))
            logger.trace("dye:{}".format([var for _, var in board(mode="dye", key=key)]))

        raise ModelGenerateError("model Error")

    if logger.print_level <= logger.TRACE:
        if not os.path.exists(CONFIG["output_path"]):
            os.makedirs(CONFIG["output_path"])
        model.ExportToFile("log/model.txt")

    if status not in (cp_model.OPTIMAL, cp_model.FEASIBLE):
        logger.trace(f"求解器status:{status}")
        return 0

    if bool_mode:
        # 如果打开 那么就将状态压缩至 有解/无解
        return 1

    current_solution = []
    for key in board.get_interactive_keys():
        for pos, _ in board("N", key=key):
            var = board.get_variable(pos)
            val = solver.Value(var)
            if val == 1:
                current_solution.append(var.Not())  # 这个变量不能为1，取反
            else:
                current_solution.append(var)  # 这个变量不能为0，保持原变量

    model.AddBoolOr(current_solution)  # 新增排除当前解约束（至少有一个变量取反）

    status2 = solver.Solve(model)
    if status2 == cp_model.FEASIBLE or status2 == cp_model.OPTIMAL:
        logger.trace(f"求解器多解")
        return 2
    return 1


def deduced_by_csp(
        board: AbstractBoard,
        answer_board: AbstractBoard,
        pos: AbstractPosition,
):
    """
    检查是否无解
    """
    if board[pos] is not None:
        return None
    model = board.get_model().clone()
    model: cp_model.CpModel

    target_var = board.get_variable(pos)
    model.Add(target_var == (0 if answer_board.get_type(pos) == "F" else 1))

    solver = get_solver(False)
    state = solver.Solve(model)

    if state == cp_model.INFEASIBLE:
        return True
    elif state in (cp_model.OPTIMAL, cp_model.FEASIBLE):
        return False
    else:
        return None


def hint_by_csp(
    board: AbstractBoard,
    answer_board: AbstractBoard,
    switch: Switch,
    pos: AbstractPosition,
    upper_bound=None
):
    if board[pos] is not None:
        return None
    model = board.get_model().clone()
    model: cp_model.CpModel

    target_var = board.get_variable(pos)
    model.Add(target_var == (0 if answer_board.get_type(pos) == "F" else 1))
    assumptions = switch.get_all_vars()

    get_logger().trace(f"pos {pos}: start\n", end="")
    results = _hint_by_csp(
        model, assumptions,
        upper_bound, 0, pos
    )
    get_logger().trace(f"pos {pos}: {results}\n", end="")
    if results is None:
        return None
    return [switch.get_obj_and_index_by_var(r) for r in results]


def _hint_by_csp(
    model: cp_model.CpModel,
    assumptions: List[IntVar],
    upper_bound=None,
    offset=0,
    pos=None
):
    logger = get_logger()
    logger.trace(f"pos {pos} off {offset}: start\n", end="")
    future_to_param = {}
    _results = []
    with ThreadPoolExecutor(max_workers=CONFIG["workes_number"]) as executor:
        for var in assumptions:
            _model = model.clone()
            _model.Add(var == 0)
            _model.AddBoolAnd([v for v in assumptions if v != var])
            solver = get_solver(True)
            fut = executor.submit(
                solver.Solve,
                _model, None
            )
            future_to_param[fut] = var
        for fut in as_completed(future_to_param):
            try:
                var = future_to_param[fut]
                logger.trace(f"pos {pos} off {offset} wait: {var}")
                status = fut.result()
                logger.trace(f"pos {pos} off {offset} end wait: {status}")
                if status in (cp_model.FEASIBLE, cp_model.OPTIMAL):
                    _results.append(var)
            except Exception as e:
                logger.trace(e)
                continue

    logger.trace(f"pos {pos} off {offset} end AND [{_results}]")

    if upper_bound is not None and len(_results) - 2 > (upper_bound[0] - offset):
        logger.trace(f"pos {pos}, off {offset}: fail (len>ub)\n", end="")
        return None
    # 将与的状态保留至model
    model.AddBoolAnd(_results)
    # 将其他约束加入进行测试是否无解(可推)
    _model = model.clone()
    _model.Add(sum([v for v in assumptions if v not in _results]) == 0)
    solver = get_solver(False)
    logger.trace(f"pos {pos} off {offset} verification and start")
    status = solver.Solve(_model)
    logger.trace(f"pos {pos} off {offset} verification and end {status}")
    if status == cp_model.INFEASIBLE:
        # 无解说明已经是包含所有的约束了 是当前层的MUS
        if upper_bound is not None:
            with upper_bound[1]:
                if (len(_results) + offset) < upper_bound[0]:
                    upper_bound[0] = len(_results) + offset
        logger.trace(f"pos {pos} off {offset}: done {_results}\n", end="")
        return _results
    elif status not in (cp_model.OPTIMAL, cp_model.FEASIBLE):
        # 未知状况直接返回None
        logger.trace(f"pos {pos} off {offset}: error state:{status}\n", end="")
        return None
    # 有解说明当前层依旧有或节点未被发现 需要继续遍历
    # 获取当前层的MCS
    mcs = [v for v in assumptions if v not in _results]

    # 使用二分推理或关系
    R = 0
    L = len(mcs)
    mid = 2 * len(mcs) // 3

    logger.trace(f"pos {pos} off {offset}: start OR lever\n", end="")
    while True:
        _model = model.clone()

        _model.Add(sum(mcs) >= mid)

        solver = get_solver(False)
        logger.trace(f"pos {pos} off {offset}: OR => start {len(mcs)}-{L}-{mid}-{R}\n", end="")
        status = solver.Solve(_model)
        logger.trace(f"pos {pos} off {offset}: OR => end {len(mcs)}-{L}-{mid}-{R} => {status}\n", end="")
        if status == cp_model.INFEASIBLE:
            # 无解
            L = mid
            mid = (L - R) // 2 + R
        elif status in (cp_model.OPTIMAL, cp_model.FEASIBLE):
            # 有解
            R = mid
            mid = (L - R) // 2 + R
            if R == mid: break
        else: return None

    _mcs = [i for i in mcs if solver.Value(i) == 0]
    logger.trace(f"pos {pos} off {offset} ORlever: {_mcs} status {status}\n", end="")
    results = []

    for var in _mcs:
        _model = model.clone()
        # 只开启其中一个或节点进行广搜
        _model.Add(var == 1)
        _model.AddBoolAnd([v.Not() for v in _mcs if v != var])
        _assumptions = [v for v in assumptions if v not in _mcs + _results]

        result = _hint_by_csp(
            _model,
            _assumptions[:],
            upper_bound=upper_bound,
            offset=offset + len(_results) + len(_mcs),
            pos=pos
        )
        if result is None:
            # 该节点求解失败或者出现错误或者提前步出
            continue
        results.append(result + [var] + _results)

    if not results:
        logger.trace(f"pos {pos} off {offset}: fail (OR)\n", end="")
        return None
    min_length = min([len(r) for r in results])
    logger.trace(f"pos {pos} off {offset}: {results}, min:{min_length}\n", end="")
    return [result for result in results if (len(result) == min_length)][0]


def solver_board(
    board: AbstractBoard,
    all_rules: List[AbstractRule]
):
    model = board.get_model()
    switch = Switch()

    for rule in all_rules:
        if rule is None:
            continue
        rule: AbstractRule
        rule.create_constraints(board, switch)

    for key in board.get_board_keys():
        for pos, obj in board(key=key):
            if obj is None:
                continue
            obj: AbstractValue
            obj.create_constraints(board, switch)

    # 3.获取所有变量并赋值已解完的部分
    for key in board.get_board_keys():
        for _, var in board("C", mode="variable", key=key):
            model.Add(var == 0)
        for _, var in board("F", mode="variable", key=key):
            model.Add(var == 1)

    model.AddBoolAnd(switch.get_all_vars())

    solver = get_solver(True)
    status = solver.Solve(model)
    positions = []

    if status not in (cp_model.OPTIMAL, cp_model.FEASIBLE):
        return None

    for key in board.get_board_keys():
        for pos, var in board(key=key, mode="variable"):
            if (
                board.get_type(pos) == "N" and
                solver.Value(var)
            ):
                positions.append(pos)

    return positions


def solver_model(
    model: cp_model,
    back_solver: bool = False
):
    solver = get_solver(False)
    status = solver.Solve(model)
    if back_solver:
        return status in (cp_model.OPTIMAL, cp_model.FEASIBLE), solver
    return status in (cp_model.OPTIMAL, cp_model.FEASIBLE)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/19 20:34
# @Author  : Wu_RH
# @FileName: 2D.py
"""
[2D']偏移: 线索表示四方向任意偏移一格为中心的3x3区域内的总雷数
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.tool import get_logger, get_random


class Rule2Dp(AbstractClueRule):
    name = ["2D'", "四向偏移"]
    doc = "线索表示四方向任意偏移一格为中心的3x3区域内的总雷数(全局统一)"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        logger = get_logger()
        for pos, _ in board("N"):
            _pos = random.choice([pos.up(1), pos.down(1), pos.right(1), pos.left(1)])
            value = len([__pos for __pos in _pos.neighbors(0, 2) if board.get_type(__pos) == "F"])
            board.set_value(pos, Value2Dp(pos, count=value))
            logger.debug(f"Set {pos} to 2D[{value}]")
        return board


class Value2Dp(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
        else:
            # 直接初始化
            self.count = count
        self.neighbors = []
        for _pos in [pos.up(1), pos.down(1), pos.right(1), pos.left(1)]:
            self.neighbors.append(_pos.neighbors(0, 2))

    def __repr__(self):
        return f"{self.count}"

    @classmethod
    def type(cls) -> bytes:
        return Rule2Dp.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束: 周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集周围格子的布尔变量
        neighbors_vars = []
        for neighbor in self.neighbors:  # 8方向相邻格子
            neighbors_vars.append(board.batch(neighbor, mode="variable", drop_none=True))

        # 添加约束：周围雷数等于count
        var_list = []
        for neighbor_vars in neighbors_vars:
            b = model.NewBoolVar("t")
            model.Add(sum(neighbor_vars) == self.count).OnlyEnforceIf(b)
            model.Add(sum(neighbor_vars) != self.count).OnlyEnforceIf(b.Not())
            var_list.append(b)

        model.AddBoolOr(var_list).OnlyEnforceIf(s)



"""
[GL] 生命游戏 (Game of Life)：每个雷周围八格中恰有2或3个雷。
"""

from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch

class RuleGL(AbstractMinesRule):
    name = ["GL", "生命游戏", "Game of Life"]
    doc = "每个雷周围八格中恰有2或3个雷"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="var"):
            var_list = board.batch(pos.neighbors(2), mode="var", drop_none=True)
            model.Add(sum(var_list) >= 2).OnlyEnforceIf([var, s])
            model.Add(sum(var_list) <= 3).OnlyEnforceIf([var, s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/12 18:58
# @Author  : Wu_RH
# @FileName: MC.py
"""
[MC]染色格合并（Mine-Combination）:线索表示周围八格的雷数，染色格中如有雷就同一算为一雷（剩余雷数不受影响）
"""
from minesweepervariants.abs.Rrule import AbstractClueRule, AbstractClueValue
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch


class RuleMC(AbstractClueRule):
    name = ["MC", "染色格合并", "Mine-Combination"]
    doc = "线索表示周围八格的雷数，染色格中如有雷就同一算为一雷（剩余雷数不受影响）"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            nei = pos.neighbors(2)
            type_list = board.batch(nei, mode="type")
            dye_list = board.batch(nei, mode="dye")
            value = [t for t, d in zip(type_list, dye_list) if not d].count("F")
            value += [t for t, d in zip(type_list, dye_list) if d].count("F") > 0
            obj = ValueMC(pos, bytes([value]))
            board[pos] = obj
        return board


class ValueMC(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos, code)
        self.value = code[0]

    def __repr__(self) -> str:
        return str(self.value)

    @classmethod
    def type(cls) -> bytes:
        return RuleMC.name[0].encode()

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)

        nei = [_pos for _pos in self.pos.neighbors(2) if board.is_valid(_pos)]
        var_list = board.batch(nei, mode="var")
        dye_list = board.batch(nei, mode="dye")

        var1_list = [v for v, d in zip(var_list, dye_list) if not d]
        var2_list = [v for v, d in zip(var_list, dye_list) if d]

        tmp = model.NewBoolVar("tmp")
        model.Add(sum(var1_list + [tmp]) == self.value).OnlyEnforceIf(s)
        # model.AddBoolOr(var2_list).OnlyEnforceIf(tmp)
        # model.AddBoolAnd([v.Not() for v in var2_list]).OnlyEnforceIf(tmp.Not())
        model.Add(sum(var2_list) > 0).OnlyEnforceIf(tmp)
        model.Add(sum(var2_list) == 0).OnlyEnforceIf(tmp.Not())



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/14 00:33
# @Author  : Wu_RH
# @FileName: 4T.py
"""
[4T]温度计(Temperature): 没写
"""

from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch
from minesweepervariants.utils.tool import get_random


class Rule4T(AbstractMinesRule):
    # name = ["4T", "温度计", "Temperature"]
    doc = "将会随机"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        def put_temperature():
            nonlocal board, deny_map, positions
            nonlocal size, random
            end_pos = random.choice([
                _pos for _pos in positions
                if _pos not in deny_map
            ])
            index = int(random.random() * 6)
            if index % 2:
                lenght = int(random.random() * (size[0] - end_pos.x)) + end_pos.x
            else:
                lenght = int(random.random() * (size[1] + end_pos.y)) + end_pos.y

        super().__init__(board, data)
        random = get_random()
        self.map = []
        deny_map = {}
        for key in board.get_interactive_keys():
            size = board.get_config(key, "size")
            positions = [pos for pos, _ in board(key=key)]
            attr_index = 3
            for x in range(100, -1, -1):
                if not x:
                    raise ValueError("Temperature init fail")
                try:
                    put_temperature()
                except:
                    continue
                attr_index -= 1
                if attr_index == 0:
                    break

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        super().create_constraints(board, switch)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 18:09
# @Author  : Wu_RH
# @FileName: 1M.py
"""
[1M']多雷': 每个线索的多雷位置相对于线索固定 且位置全盘共享(总雷数不受限制)
"""

from ....abs.board import AbstractBoard, AbstractPosition
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....utils.tool import get_random
from ....utils.impl_obj import VALUE_CIRCLE, VALUE_CROSS

BOARD_NAME = "1M'"


class Rule1M(AbstractClueRule):
    name = ["1M'", "多雷'"]
    doc = "每个线索的多雷位置相对于线索固定 且位置全盘共享(总雷数不受限制)"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        board.generate_board(BOARD_NAME, (3, 3))
        if data is None:
            self.value = 4
        else:
            self.value = data

    def fill(self, board: 'AbstractBoard'):
        def apply_offsets(_pos: AbstractPosition):
            nonlocal offsets
            result = []
            for dpos in offsets:
                result.append(_pos.deviation(dpos))
            return result
        random = get_random()
        pos = board.get_pos(1, 1, BOARD_NAME)

        if self.value is None:
            board[pos] = Value2I_7(pos, bytes([4]))
        elif self.value == "":
            board[pos] = Value2I_7(pos, bytes([9]))
        else:
            board[pos] = Value2I_7(pos, bytes([int(self.value)]))

        pos_list = [pos for pos, _ in board("N", key=BOARD_NAME)]

        if self.value == "":
            pos_list = random.sample(pos_list, int(random.random() * 9))
        else:
            pos_list = random.sample(pos_list, int(self.value))
        offsets = []
        for pos in pos_list:
            board[pos] = VALUE_CIRCLE
            offsets.append(pos.up().left())
        for pos, _ in board("N", key=BOARD_NAME):
            board[pos] = VALUE_CROSS

        for pos, _ in board("N"):
            positions = pos.neighbors(2)
            offset_poses = apply_offsets(pos)
            value = board.batch(positions, mode="type").count("F")
            value += board.batch(offset_poses, mode="type").count("F")
            obj = Value1M(pos, code=bytes([value]))
            board.set_value(pos, obj)
        return board

    def init_clear(self, board: 'AbstractBoard'):
        for pos, obj in board(mode="object", key=BOARD_NAME):
            if isinstance(obj, Value2I_7):
                continue
            board[pos] = None


class Value1M(AbstractClueValue):
    value: int
    neighbors: list

    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        self.value = code[0]
        self.neighbors = pos.neighbors(2)

    def __repr__(self) -> str:
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        positions = self.neighbors[:]
        neighbors = []
        for pos2, obj in board(key=BOARD_NAME):
            if isinstance(obj, Value2I_7):
                continue
            neighbors.append([self.pos.deviation(pos2).up().left(), pos2])
        for pos, pos2 in neighbors:
            if board.get_type(pos) == "F":
                positions.append(pos2)
        return positions

    @classmethod
    def type(cls) -> bytes:
        return Rule1M.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        var_list = board.batch(self.neighbors, mode="variable", drop_none=True)

        # 初始化对照表
        neighbors = []
        for pos2, obj in board(key=BOARD_NAME):
            if isinstance(obj, Value2I_7):
                continue
            # 题板上的位置和共享的偏移位置
            _positions = [self.pos.deviation(pos2).up().left(), pos2]
            # 第一个为题板对应的变量 第二个为偏移的变量
            neighbors.append(board.batch(_positions, mode="variable"))

        # 初始化和值
        for var_to_sum, cond in neighbors:
            if var_to_sum is None or cond is None:
                continue
            # 初始化临时变量
            tmp = model.NewBoolVar(f"included_if_{self.pos}_{var_to_sum}")
            # 如果偏移变量为真 那么tmp为题板的值
            model.Add(tmp == var_to_sum).OnlyEnforceIf([cond, s])
            # 如果偏移变量为假 那么tmp为0
            model.Add(tmp == 0).OnlyEnforceIf([cond.Not(), s])
            var_list.append(tmp)

        model.Add(sum(var_list) == self.value).OnlyEnforceIf(s)


class Value2I_7(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos, code)
        self.neighbors = pos.neighbors(2)
        self.value = code[0]

    def __repr__(self) -> str:
        return str(self.value) if self.value < 9 else "?"

    @classmethod
    def type(cls) -> bytes:
        return Rule1M.name[0].encode("ascii") + b"_n"

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        if self.value > 8:
            return
        model = board.get_model()
        s = switch.get(model, self)
        model.Add(sum(board.batch(self.neighbors, mode="variable")) == self.value).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/07 16:08
# @Author  : Wu_RH
# @FileName: 3D1K.py
from typing import List

from minesweepervariants.abs.Rrule import AbstractClueValue
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from .. import Abstract3DClueRule


class Rule3D1K(Abstract3DClueRule):
    name = ["3DK", "3D1K", "3D标准扫雷"]
    doc = "每个数字标明周围26格内雷的数量。"

    def __init__(self, board: AbstractBoard, data: str = None):
        super().__init__(board, data)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            positions = self.pos_neighbors(board, pos, 5, 5)
            value = board.batch(positions, mode="type").count("F")
            board.set_value(pos, ValueK(pos, code=bytes([value])))
        return board


class ValueK(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.pos = pos

    def __repr__(self):
        return f"{self.value}"

    @classmethod
    def type(cls) -> bytes:
        return Rule3D1K.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        return Rule3D1K.pos_neighbors(board, self.pos, 5, 5)

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束: 周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        neighbor = Rule3D1K.pos_neighbors(board, self.pos, 5, 5)

        # 收集周围格子的布尔变量
        neighbor_vars = []
        for neighbor in neighbor:  # 8方向相邻格子
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：周围雷数等于count
        if neighbor_vars:
            model.Add(sum(neighbor_vars) == self.value).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/28 12:39
# @Author  : Wu_RH
# @FileName: 4T.py
"""
[*3T]:雷线索指示包含自身的雷三连数量。雷三连允许部分重合
"""
from ....abs.Mrule import AbstractMinesClueRule, AbstractMinesValue
from ....abs.board import AbstractPosition, AbstractBoard

COUNT = 0


class Rule4T(AbstractMinesClueRule):
    name = ["*3T", "雷三连"]
    doc = "雷线索指示包含自身的雷三连数量。雷三连允许部分重合"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for _pos, _ in board("F"):
            board.set_value(_pos, Value4T(_pos))
        for _pos, _ in board("F"):
            positions = [
                [_pos, _pos.left(), _pos.left(2)],
                [_pos, _pos.down(), _pos.down(2)],
                [_pos, _pos.left().down(), _pos.left(2).down(2)],
                [_pos, _pos.right().down(), _pos.right(2).down(2)],
            ]
            for poses in positions:
                if board.batch(poses, "type").count("F") != 3:
                    continue
                for pos in poses:
                    board[pos].value += 1
        return board

    def create_constraints(self, board: 'AbstractBoard', switch):
        global COUNT
        COUNT += 1
        model = board.get_model()
        c_map = {}
        r_map = {}
        d1_map = {}
        d2_map = {}
        map_list = [c_map, r_map, d1_map, d2_map]
        for _pos, _ in board():
            positions = [
                [_pos, _pos.left(), _pos.left(2)],
                [_pos, _pos.down(), _pos.down(2)],
                [_pos, _pos.left().down(), _pos.left(2).down(2)],
                [_pos, _pos.right().down(), _pos.right(2).down(2)],
            ]
            for i in range(4):
                var_list = board.batch(positions[i], "variable")
                if any(v is None for v in var_list):
                    continue
                t = model.NewBoolVar("[*3T]")
                model.Add(sum(var_list) == 3).OnlyEnforceIf(t)
                model.Add(sum(var_list) < 3).OnlyEnforceIf(t.Not())
                map_list[i][_pos] = t

        for _pos, obj in board("F"):
            if type(obj) is not Value4T:
                continue
            positions = [
                [_pos, _pos.right(), _pos.right(2)],
                [_pos, _pos.up(), _pos.up(2)],
                [_pos, _pos.right().up(), _pos.right(2).up(2)],
                [_pos, _pos.left().up(), _pos.left(2).up(2)],
            ]
            var_list = []
            for i in range(4):
                for pos in positions[i]:
                    if pos not in map_list[i]:
                        continue
                    var_list.append(map_list[i][pos])
            obj.create_constraints_(model, var_list, switch.get(model, obj))


class Value4T(AbstractMinesValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = None):
        self.value = code[0] if code else 0
        self.pos = pos

    def __repr__(self):
        return str(self.value)

    def code(self) -> bytes:
        return bytes([self.value])

    @classmethod
    def type(cls) -> bytes:
        return Rule4T.name[0].encode("ascii")

    def create_constraints_(self, model, var_list: list, s):
        model.Add(sum(var_list) == self.value).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/21 05:57
# @Author  : Wu_RH
# @FileName: CCB.py
"""
[CB]连击（Combo）: 雷线索指示四方向的连击总值和，同方向相邻的n个雷连击值为1-n.例：FXFFFXX这一支连击总值为7
"""
from minesweepervariants.abs.Mrule import AbstractMinesValue, AbstractMinesClueRule
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch


class RuleCB(AbstractMinesClueRule):
    name = ["CB", "CCB", "连击"]
    doc = "雷线索指示四方向的连击总值和，同方向相邻的n个雷连击值为1-n"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, obj in board("F"):
            col = board.get_col_pos(pos)
            row = board.get_row_pos(pos)
            col1, col2 = col[:col.index(pos)][::-1], col[col.index(pos)+1:]
            row1, row2 = row[:row.index(pos)][::-1], row[row.index(pos)+1:]
            value = 0
            for line in [row1, row2, col1, col2]:
                line = [t == "F" for t in board.batch(line, mode="type")]
                tmp = 0
                for b in line:
                    if b:
                        tmp += 1
                        value += tmp
                    else:
                        tmp = 0
            obj = ValueCB(pos, bytes([value]))
            board[pos] = obj
        return board


class ValueCB(AbstractMinesValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = None):
        self.pos = pos
        self.value = code[0]

    def __repr__(self):
        return str(self.value)

    def code(self) -> bytes:
        return bytes([self.value])

    @classmethod
    def type(cls) -> bytes:
        return RuleCB.name[0].encode()

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)

        col = board.get_col_pos(self.pos)
        row = board.get_row_pos(self.pos)

        col1, col2 = col[:col.index(self.pos)], col[col.index(self.pos) + 1:]
        row1, row2 = row[:row.index(self.pos)], row[row.index(self.pos) + 1:]

        var_list = []
        for pos_line in [col1, col2, row1, row2]:
            if not pos_line:
                continue
            var_line = board.batch(pos_line, mode="var")
            type_line = board.batch(pos_line, mode="type")
            for i in range(1, len(pos_line)+1):
                for index in range(len(pos_line) - i + 1):
                    tmp_var_line = var_line[index:index+i]
                    tmp_type_line = type_line[index:index+i]
                    if any(v is None for v in tmp_var_line):
                        continue
                    if "C" in tmp_type_line:
                        continue
                    tmp_bool_var = model.NewBoolVar(f"{pos_line[index]}:{pos_line[index+i-1]}")
                    for v in tmp_var_line:
                        model.Add(tmp_bool_var == 0).OnlyEnforceIf(v.Not())
                    model.Add(tmp_bool_var == 1).OnlyEnforceIf(tmp_var_line)
                    var_list.append(tmp_bool_var)
        model.Add(sum(var_list) == self.value).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/03 11:37
# @Author  : Wu_RH
# @FileName: 2I.py
"""
[2I]残缺：数字表示周围8格中某7格的雷数。7格的方位被当前题板所有线索共享
"""

from ....abs.Rrule import AbstractClueValue, AbstractClueRule
from ....abs.board import AbstractPosition, AbstractBoard
from ....utils.impl_obj import VALUE_CROSS, VALUE_CIRCLE
from ....utils.tool import get_random, get_logger

NAME_2I = "2I"


class Rule2I(AbstractClueRule):
    name = ["2I", "残缺"]
    doc = "数字表示周围8格中某7格的雷数。7格的方位被当前题板所有线索共享"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        board.generate_board(NAME_2I, (3, 3))

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        def apply_offsets(_pos: AbstractPosition):
            nonlocal offsets
            result = []
            for dpos in offsets:
                result.append(_pos.deviation(dpos))
            return result

        random = get_random()
        logger = get_logger()
        pos = board.get_pos(1, 1, NAME_2I)
        board[pos] = Value2I_7(pos)

        pos_list = [pos for pos, _ in board("N", key=NAME_2I)]
        pos_list = random.sample(pos_list, 7)
        offsets = []
        for pos in pos_list:
            board[pos] = VALUE_CIRCLE
            offsets.append(pos.up().left())
            logger.debug(f"[2I] put O at {pos}")
        for pos, _ in board("N", key=NAME_2I):
            board[pos] = VALUE_CROSS
            logger.debug(f"[2I] put X at {pos}")

        for pos, _ in board("N"):
            positions = apply_offsets(pos)
            value = board.batch(positions, mode="type", drop_none=True).count("F")
            obj = Value2I(pos, bytes([value]))
            board.set_value(pos, obj)
            logger.debug(f"[2I] put ({value}) at {pos}")

        return board

    def init_clear(self, board: 'AbstractBoard'):
        for pos, obj in board(mode="object", key=NAME_2I):
            if isinstance(obj, Value2I_7):
                continue
            board[pos] = None


class Value2I(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.pos = pos
        self.value = code[0]

    def __repr__(self):
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        positions = []
        for pos, _ in board("NF", key=NAME_2I):
            _pos = self.pos.deviation(pos.shift(1, -1))
            if board.in_bounds(_pos):
                positions.append(_pos)
        return positions

    @classmethod
    def method_choose(cls) -> int:
        return 1

    @classmethod
    def type(cls) -> bytes:
        return Rule2I.name[0].encode("ascii")

    def code(self):
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        logger = get_logger()

        # 初始化对照表
        neighbors = []
        for pos2, obj in board(key=NAME_2I):
            if isinstance(obj, Value2I_7):
                continue
            # 题板上的位置和共享的偏移位置
            _positions = [self.pos.deviation(pos2).up().left(), pos2]
            # 第一个为题板对应的变量 第二个为偏移的变量
            neighbors.append(board.batch(_positions, mode="variable"))

        # 初始化和值
        sum_vers = []
        for var_to_sum, cond in neighbors:
            if var_to_sum is None or cond is None:
                continue
            # 初始化临时变量
            tmp = model.NewBoolVar(f"included_if_{self.pos}_{var_to_sum}")
            # 如果偏移变量为真 那么tmp为题板的值
            model.Add(tmp == var_to_sum).OnlyEnforceIf([cond, s])
            # 如果偏移变量为假 那么tmp为0
            model.Add(tmp == 0).OnlyEnforceIf([cond.Not(), s])
            sum_vers.append(tmp)
            logger.trace(f"[2E'2I] new tempVar: {tmp} = if {cond} -> {var_to_sum}")

        model.Add(sum(sum_vers) == self.value).OnlyEnforceIf(s)


class Value2I_7(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos, code)
        self.neighbors = pos.neighbors(2)

    def __repr__(self) -> str:
        return "7"

    @classmethod
    def method_choose(cls) -> int:
        return 1

    @classmethod
    def type(cls) -> bytes:
        return Rule2I.name[0].encode("ascii") + b"_7"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        model.Add(sum(board.batch(self.neighbors, mode="variable")) == 7).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/11 14:25
# @Author  : xxx
# @FileName: 1D.py
"""
[SM]神秘: 这是一条会让男娘感到困惑的规则
"""

from ....abs.Lrule import AbstractMinesRule


class Rule1D(AbstractMinesRule):
    name = ["SM", "???", "神秘"]
    doc = "这是一条会让男娘感到困惑的规则"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/28 21:26
# @Author  : Wu_RH
# @FileName: 1L^.py
"""
[2*1L]2倍误差：所有线索均比真实值大2或小2或者不变
"""


from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.tool import get_logger, get_random


class Rule1Lo(AbstractClueRule):
    name = ["2*1L", "2倍误差"]
    doc = "所有线索均比真实值大2或小2或者不变"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        logger = get_logger()
        for pos, _ in board("N"):
            value = len([_pos for _pos in pos.neighbors(2) if board.get_type(_pos) == "F"])
            random_number = random.random()
            if value == 0:
                value = 0 if random_number < 0.5 else 2
            elif value == 1:
                value = 1 if random_number < 0.5 else 3
            elif value == 7:
                value = 7 if random_number < 0.5 else 5
            elif value == 8:
                value = 8 if random_number < 0.5 else 6
            else:
                value += 2 if random_number > 0.6 else (0 if random_number < 0.3 else -2)
            board.set_value(pos, Value1Lo(pos, count=value))
            logger.debug(f"Set {pos} to 1L^[{value}]")
        return board


class Value1Lo(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
        else:
            # 直接初始化
            self.count = count
        self.neighbor = self.pos.neighbors(2)

    def __repr__(self):
        return f"{self.count}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule1Lo.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束：周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集周围格子的布尔变量
        neighbor_vars = []
        for neighbor in self.neighbor:  # 8方向相邻格子
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：周围雷数等于count+-1
        if neighbor_vars:
            neighbor_sum = sum(neighbor_vars)
            # 两个布尔变量表示加和为 count + 1 或 count - 1
            b1 = model.NewBoolVar("sum_eq_count_plus_2")
            b2 = model.NewBoolVar("sum_eq_count")
            b3 = model.NewBoolVar("sum_eq_count_minus_2")

            # 将布尔变量与表达式绑定
            model.Add(neighbor_sum == self.count + 2).OnlyEnforceIf([b1, s])
            model.Add(neighbor_sum != self.count + 2).OnlyEnforceIf([b1.Not(), s])

            model.Add(neighbor_sum == self.count).OnlyEnforceIf([b2, s])
            model.Add(neighbor_sum != self.count).OnlyEnforceIf([b2.Not(), s])

            model.Add(neighbor_sum == self.count - 2).OnlyEnforceIf([b3, s])
            model.Add(neighbor_sum != self.count - 2).OnlyEnforceIf([b3.Not(), s])

            model.AddBoolOr([b1, b2, b3]).OnlyEnforceIf(s)
